# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
import smtplib
from email.message import EmailMessage
import re


from .constants import Email


class EmailNotification(object):

    def __init__(self, **params):
        self._port = params.get(Email.SMTP_PORT)  
        self._smtp_server = params.get(Email.SMTP_ENDPOINT)
        self._sender_email = params.get(Email.SEND_FROM)
        self._receiver_email = params.get(Email.SEND_TO) 
        self._password = params.get(Email.SEND_FROM_PASSWORD)
        self._context = params.get(Email.CONTEXT)

        self._validate_email_params()


    @property
    def send_to(self):
        return self._receiver_email

    @property
    def context(self):
        return self._context or 'unknwon'


    def send_email(self, subject, message):
        # context = ssl.create_default_context()
        # with smtplib.SMTP_SSL(smtp_server, port, context=context) as server:

        with smtplib.SMTP(self._smtp_server, self._port) as server:
            server.starttls() 
            server.login(self._sender_email, self._password)
            result = server.sendmail(self._sender_email, self._receiver_email, 
                                     f"From: {self._sender_email}\nTo: {self._receiver_email}\nSubject: {subject}\n\nContext: {self.context}\n{message}")
            if result:
                error_message = []
                for key, value in result:
                    error_message.append(f"failed to send mail to {key} due to error {value}")
                raise ValueError(",\n".join(error_message))


    def _validate_email_params(self):
        if self._port and self._smtp_server and self._sender_email and self._receiver_email and self._password:
            if self._is_email_format(self._sender_email) and self._is_email_format(self._receiver_email):
                return True
        raise ValueError("""
            cannot notify device_code login by email because some email parameters are missing. 
            Set KQLMAGIC_CODE_NOTIFICATION_EMAIL in the following way: SMTPEndPoint: \" email server\"; SMTPPort: \"email port\";
            sendFrom: \"sender email address \"; sendFromPassword: \"email address password \"; sendTo:\" email address to send to\"""")


    def _is_email_format(self, email):
        return re.match(r'[\w\.-]+@[\w\.-]+(\.[\w]+)+', email)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------


from .ipython_api import IPythonAPI


class Help_html(object):
    """
    adds entries to jupyter help
    """

    showfiles_base_url = None
    _pending_helps = {}


    @staticmethod
    def flush(window_location:str, options:dict=None):
        options = options or {}
        base_url = None
        # local machine jupyter
        if window_location.startswith(("http://localhost", "https://localhost", "http://127.0.0.", "https://127.0.0.")):           
            # example: http://localhost:8888/notebooks/my%20notebooks/legend.ipynb
            parts = window_location.split("/")
            parts.pop()  # remove notebook name 
            del parts[3]  # remove '/notebooks'
            base_url = "/".join(parts)

        # know remote service address
        else:
            notebook_service_address = options.get("notebook_service_address")
            if notebook_service_address is not None:
                # azure notebooks service
                if notebook_service_address.endswith("notebooks.azure.com"):
                    start = notebook_service_address.find("//") + 2
                    host_suffix = "." + notebook_service_address[start:]

                    start = window_location.find("//")
                    end = window_location.find(host_suffix)
                    # azure notebook environment, assume template: https://library-user.notebooks.azure.com
                    if start > 0 and end > 0 and ('-' in window_location):
                        library_user_segment = window_location[start + 2: end]
                        if '-' in library_user_segment:
                            library, user = library_user_segment.split("-", 1)
                            base_url = f"{notebook_service_address}/api/user/{user}/library/{library}/html"

                # other remote services (assume support /tree)
                if base_url is None:
                    parts = notebook_service_address.split("/")
                    base_url = "/".join(parts[:3])  + '/tree'
                
            # default (assume support /tree)
            if base_url is None:
                # assume a remote kernel url pattern  as local
                parts = window_location.split("/")
                base_url = "/".join(parts[:3])  + '/tree'

        Help_html.showfiles_base_url = base_url
        refresh = False
        for text, url in Help_html._pending_helps.items():
            Help_html.add_menu_item(text, url, False, **options)
            refresh = True
        Help_html._pending_helps = {}
        if refresh:
            IPythonAPI.try_kernel_reconnect(**options)


    @staticmethod
    def add_menu_item(text, file_path: str, reconnect=True, **options):
        if not text:
            return

        if not file_path:
            return

        # add help link
        if file_path.startswith("http"):
            url = file_path
        elif Help_html.showfiles_base_url is not None:
            url = f"{Help_html.showfiles_base_url}/{file_path}"
        else:
            url = None

        if url:
            IPythonAPI.try_add_to_help_links(text, url, reconnect, **options)
        elif Help_html._pending_helps.get(text) is None:
            Help_html._pending_helps[text] = file_path



#!/usr/bin/python
# -*- coding: utf-8 -*-
# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

# --------------------------------------------------------------------------
# because it is also executed from setup.py, make sure
# that it imports only modules, that for sure will exist at setup.py execution
# TRY TO KEEP IMPORTS TO MINIMUM
# --------------------------------------------------------------------------

# Union Without repetition  
def list_union(*args): 
    final_list = list(set().union(*args)) 
    return final_list

def strip_package_name(item):
    "parse package name from a package line that include version info"
    item = item.strip().lower()
    for i in range(len(item)):
        if item[i] in ' <>=~!;':
           return item[:i]
    return item

# ---------------------------------------------------
_IPYTHON_REQUIRES    = [
    'ipython>=7.1.1',
]

_JUPYTER_REQUIRES    = [ 
    'ipykernel>=5.1.1',
]
# ---------------------------------------------------


# ---------------------------------------------------
_ONLINE_REQUIRES     = [ 
    'requests>=2.21.0',
]

_AUTH_REQUIRES = [
    'msal>=1.11.0',
]

_PRETTY_TABLE_RQUIRES = [ 
    'prettytable>=0.7.2',    
]

_BASIC_REQUIRES = list_union(_ONLINE_REQUIRES, _AUTH_REQUIRES, _PRETTY_TABLE_RQUIRES)
# ---------------------------------------------------


# ---------------------------------------------------
_BASE_UTILS_REQUIRES = [
    'psutil>=5.4.7',
    'setuptools>=41.0.1', # optional  
    'flask>=1.0.3', # must have for Azure Data Studio SAW
    'isodate>=0.6.0',
]

_TEXT_UTILS_REQUIRES = [
    'Markdown>=3.0.1',
    'beautifulsoup4>=4.6.3',
    'lxml>=4.2.5',
]

_JSON_COLOR_REQUIRES = [
    'Pygments>=2.2.0',
]

_AUTH_CODE_TO_CLIPBOARD_REQUIRES = [
    'pyperclip>=1.7.0',
]

_MATPLOTLIB_PALETTES_REQUIRES = [
    'matplotlib>=3.0.0',
]

_UTILS_REQUIRES = list_union(_BASE_UTILS_REQUIRES, _TEXT_UTILS_REQUIRES, _JSON_COLOR_REQUIRES, _AUTH_CODE_TO_CLIPBOARD_REQUIRES,
                               _MATPLOTLIB_PALETTES_REQUIRES)
# ---------------------------------------------------


# ---------------------------------------------------
_PLOT_PLOTLY_REQUIRES = [
    'plotly>=3.10.0',
]

_PLOT_REQUIRES = list_union(_PLOT_PLOTLY_REQUIRES)

_DATAFRAME_PANDAS_REQUIRES = [
    'pandas>=0.23.4',
]

_DATAFRAME_REQUIRES = list_union(_DATAFRAME_PANDAS_REQUIRES)

_IPYWIDGETS_REQUIRES = [
    'ipywidgets',
]

_KQLMAGIC_FERNET_SSO_REQUIRES = [
    'cryptography>=2.7', # also required by adal
    'password-strength>=0.0.3',
]

_KQLMAGIC_MSAL_SSO_REQUIRES = [
    'msal-extensions>=0.3.0', # required for SSO
]


_KQLMAGIC_SSO_REQUIRES = list_union(_AUTH_REQUIRES, _KQLMAGIC_MSAL_SSO_REQUIRES, _KQLMAGIC_FERNET_SSO_REQUIRES)

_AZCLI_SSO_REQUIRES = [
    'azure-identity>=1.5.0',
    'azure-common>=1.1.25',
]

_VSCODE_SSO_REQUIRES = [
    'azure-identity>=1.5.0',
]

_MSAL_SSO_REQUIRES = [
    'msal-extensions>=0.3.0', # required for SSO
]

_MSI_SSO_REQUIRES = [
    'msrestazure>=0.6.3',
]

_SSO_REQUIRES = list_union(_KQLMAGIC_SSO_REQUIRES, _AZCLI_SSO_REQUIRES, _VSCODE_SSO_REQUIRES, _MSAL_SSO_REQUIRES,
                          _MSI_SSO_REQUIRES)

_EXTRA_REQUIRES = list_union(_PLOT_REQUIRES, _DATAFRAME_REQUIRES, _SSO_REQUIRES, _IPYWIDGETS_REQUIRES)
# ---------------------------------------------------

_SAW_REQUIRES = list_union(_JUPYTER_REQUIRES, _IPYTHON_REQUIRES, _IPYWIDGETS_REQUIRES, _DATAFRAME_PANDAS_REQUIRES, _AUTH_REQUIRES, _JSON_COLOR_REQUIRES)


# the most slim configuration
_NAKED_REQUIRES = []

# includes all configuration

_PYTHON_BASIC_REQUIRES  = list_union(_BASIC_REQUIRES)
_JUPYTER_BASIC_REQUIRES = list_union(_PYTHON_BASIC_REQUIRES, _JUPYTER_REQUIRES)
_IPYTHON_BASIC_REQUIRES = list_union(_PYTHON_BASIC_REQUIRES, _IPYTHON_REQUIRES)
_BASIC_REQUIRES         = list_union(_PYTHON_BASIC_REQUIRES, _JUPYTER_REQUIRES, _IPYTHON_REQUIRES)

_PYTHON_EXTENDED_REQUIRES  = list_union(_PYTHON_BASIC_REQUIRES, _UTILS_REQUIRES, _SSO_REQUIRES)
_JUPYTER_EXTENDED_REQUIRES = list_union(_PYTHON_EXTENDED_REQUIRES, _JUPYTER_REQUIRES)
_IPYTHON_EXTENDED_REQUIRES = list_union(_PYTHON_EXTENDED_REQUIRES, _IPYTHON_REQUIRES)
_EXTENDED_REQUIRES         = list_union(_PYTHON_EXTENDED_REQUIRES, _JUPYTER_REQUIRES, _IPYTHON_REQUIRES)

_PYTHON_ALL_REQUIRES  = list_union(_PYTHON_EXTENDED_REQUIRES, _EXTRA_REQUIRES)
_JUPYTER_ALL_REQUIRES = list_union(_PYTHON_ALL_REQUIRES, _JUPYTER_REQUIRES)
_IPYTHON_ALL_REQUIRES = list_union(_PYTHON_ALL_REQUIRES, _IPYTHON_REQUIRES)
_ALL_REQUIRES         = list_union(_PYTHON_ALL_REQUIRES, _JUPYTER_REQUIRES, _IPYTHON_REQUIRES)

_DEFAULT_REQUIRES = _ALL_REQUIRES

# ---------------------------------------------------
# ---------------------------------------------------
# ---------------------------------------------------
INSTALL_REQUIRES  = [
    'python-dateutil>=2.7.5', # also required by adal, pandas
    'traitlets>=4.3.2', # must have, basic
]

EXTRAS_REQUIRE      = {
    'default': _DEFAULT_REQUIRES,

    'naked': _NAKED_REQUIRES,

    'saw': _SAW_REQUIRES,

    'jupyter-basic': _JUPYTER_BASIC_REQUIRES,
    'ipython-basic': _IPYTHON_BASIC_REQUIRES,
    'python-basic': _PYTHON_BASIC_REQUIRES,
    'basic': _BASIC_REQUIRES,

    'jupyter-extended': _JUPYTER_EXTENDED_REQUIRES,
    'ipython-extended': _IPYTHON_EXTENDED_REQUIRES,
    'python-extended': _PYTHON_EXTENDED_REQUIRES,
    'extended': _EXTENDED_REQUIRES,

    'jupyter-all': _JUPYTER_ALL_REQUIRES,
    'ipython-all': _IPYTHON_ALL_REQUIRES,
    'python-all': _PYTHON_ALL_REQUIRES,

    'plotly': _PLOT_PLOTLY_REQUIRES,
    'pandas': _DATAFRAME_PANDAS_REQUIRES,
    'widgets': _IPYWIDGETS_REQUIRES,

    'sso': _SSO_REQUIRES,
    'azcli_sso': _AZCLI_SSO_REQUIRES,
    'msi_sso': _MSI_SSO_REQUIRES,
    'vscode_sso': _VSCODE_SSO_REQUIRES,
    'msal_sso': _MSAL_SSO_REQUIRES,
    'kqlmagic_sso': _KQLMAGIC_SSO_REQUIRES,
    'kqlmagic_msal_sso': _KQLMAGIC_FERNET_SSO_REQUIRES,
    'kqlmagic_fernet_sso': _KQLMAGIC_FERNET_SSO_REQUIRES,

    'utils': _UTILS_REQUIRES,
    'base_utils': _BASE_UTILS_REQUIRES,
    'text_utils': _TEXT_UTILS_REQUIRES,
    'json_color': _JSON_COLOR_REQUIRES,
    'auth_code_clipboard': _AUTH_CODE_TO_CLIPBOARD_REQUIRES,
    'matplotlib_palettes': _MATPLOTLIB_PALETTES_REQUIRES,
}

TESTS_REQUIRE       = [
    'pytest',
    'pytest-pep8',
    'pytest-docstyle',
    'pytest-flakes',
    'pytest-cov',

    'QtPy',
    'PyQt5'
]

DEV_REQUIRES       = [
    'twine',      # Collection of utilities for publishing packages on PyPI
    'pip',        # The PyPA recommended tool for installing Python packages
    'wheel',      # A built-package format for Python
    'black',      # code formatter

    'flake8',     # analyzes programs and detects various errors
    'pipreqs',    # Pip requirements.txt generator based on imports in project

    'mccabe',     # Ned’s script to check McCabe complexity, plugin for flake8
    'pep8'        # heck PEP-8 naming conventions, plugin for flake8
    'autopep8',   # A tool that automatically formats Python code to conform to the PEP 8 style guide
    'pydocstyle', # tatic analysis tool for checking compliance with Python docstring conventions
]



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

try:
    # to avoid this warnig to be dispayed:
    # UserWarning: Distutils was imported before Setuptools. This usage is discouraged and may exhibit undesirable behaviors or errors. Please use Setuptools' objects directly or at least import Setuptools first.
    import setuptools
except:
    pass

from .magic_extension import load_ipython_extension, unload_ipython_extension, _register_kqlmagic_magic
from ._version import __version__
from .kql_magic import kql, kql_stop

__all__ = ['__version__', 'kql', 'kql_stop', 'load_ipython_extension', 'unload_ipython_extension', '_register_kqlmagic_magic']



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

from traitlets import TraitError

from .constants import Constants
from .log import logger
from .kql_magic import Kqlmagic
from .commands_table import COMMANDS_TABLE


def load_ipython_extension(ip):
    """Load the extension in Jupyter."""

    # this fails in both Firefox and Chrome for OS X.
    # I get the error: TypeError: IPython.CodeCell.config_defaults is undefined

    # js = "IPython.CodeCell.config_defaults.highlight_modes['magic_kql'] = {'reg':[/^%%kql/]};"
    # display_javascript(js, raw=True)
    logger().debug("load_ipython_extension - start")
    
    if not Kqlmagic.is_kqlmagic_kernel:
        result = _register_kqlmagic_magic(ip)
    else:
        result = None
        logger().debug(f"load_ipython_extension - cannot register {Constants.MAGIC_NAME} because magic is already by kqlmagic_kernel")

        
    logger().debug("load_ipython_extension - end")
    return result


def _register_kqlmagic_magic(ip, is_kqlmagic_kernel=False):
    """Register Kqlmagic magic in Jupyter."""

    if not Kqlmagic.is_kqlmagic_kernel:
        Kqlmagic._is_kernel_init = True
        logger().debug(f"set_kqlmagic_magic - register magic {Constants.MAGIC_NAME}")
        # must be before registration
        Kqlmagic.is_ipython_extension = True
        Kqlmagic.is_kqlmagic_kernel = is_kqlmagic_kernel
        
        # register magic, also create magic object
        result = ip.register_magics(Kqlmagic)       
        for alias in Constants.MAGIC_ALIASES:
            logger().debug(f"register_kqlmagic_magic - register '{alias}' cell alias for {Constants.MAGIC_NAME}")
            ip.magics_manager.register_alias(alias, Constants.MAGIC_NAME, "cell")

            logger().debug(f"register_kqlmagic_magic - register '{alias}' line alias for {Constants.MAGIC_NAME}")        
            ip.magics_manager.register_alias(alias, Constants.MAGIC_NAME, "line")

        if is_kqlmagic_kernel:
            _set_command_magics(ip)

    else:
        result = None

    return result


def _set_command_magics(ip):
    try:
        for cmd_name in COMMANDS_TABLE:
            flag:str = COMMANDS_TABLE.get(cmd_name).get("flag")
            is_cell:bool = COMMANDS_TABLE.get(cmd_name).get("cell")
            _register_command_magics(ip, flag, is_cell)
    except:
        pass


def _register_command_magics(ip, flag, is_cell):
    kind = "line"
    _register_command_kind_magics(ip, flag, kind)

    if is_cell:
        kind = "cell"
        _register_command_kind_magics(ip, flag, kind)


def _register_command_kind_magics(ip, flag, kind, params=None):
    params = params or f"--{flag}"

    _register_command_magic(ip, flag, params, kind)
    alias = flag.replace("_", "")
    if alias != flag:
        _register_command_magic(ip, alias, params, kind)
        alias = flag.replace("_", "-")
        _register_command_magic(ip, alias, params, kind)


def _register_command_magic(ip, alias:str, params, kind):
    try:
        logger().debug(f"_register_command_magic - register alias '{alias}' {kind} alias for {Constants.MAGIC_NAME} {params}")
        ip.magics_manager.register_alias(alias, Constants.MAGIC_NAME, magic_kind=kind, magic_params=params)

    except TraitError:
        logger().debug(f"_register_command_magic - failed to register alias '{alias}' {kind} alias for {Constants.MAGIC_NAME} {params}")
        if not alias.startswith("kql_"):
            alias = f"kql_{alias}"
            _register_command_kind_magics(ip, alias, kind, params)
        else:
            logger().debug(f"_register_command_magic - failed to register alias '{alias}' {kind} alias for {Constants.MAGIC_NAME} {params}")
    except:
        logger().debug(f"_register_command_magic - failed to register alias '{alias}' {kind} alias for {Constants.MAGIC_NAME} {params}")


def unload_ipython_extension(ip):
    """Unoad the extension in Jupyter."""
    logger().debug("unload_ipython_extension - start")
    
    if not Kqlmagic.is_kqlmagic_kernel:
        Kqlmagic.stop(unload_ipython_extension=True)
        Kqlmagic.is_ipython_extension = False
        logger().debug(f"unload_ipython_extension - remove {Constants.MAGIC_NAME} from cell magics")
        del ip.magics_manager.magics["cell"][Constants.MAGIC_NAME]

        logger().debug(f"unload_ipython_extension - remove {Constants.MAGIC_NAME} from line magics")
        del ip.magics_manager.magics["line"][Constants.MAGIC_NAME]

        for alias in Constants.MAGIC_ALIASES:
            logger().debug(f"unload_ipython_extension - remove '{alias}' cell alias for {Constants.MAGIC_NAME}")
            del ip.magics_manager.magics["cell"][alias]

            logger().debug(f"unload_ipython_extension - remove '{alias}' line alias for {Constants.MAGIC_NAME}")
            del ip.magics_manager.magics["line"][alias]
    else:
        logger().debug(f"unload_ipython_extension - cannot remove {Constants.MAGIC_NAME} because magic is used by kqlmagic_kernel")
            
    logger().debug("unload_ipython_extension - end")



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

from typing import Dict
import re
import uuid
import json


from .my_aad_helper_msal import _MyAadHelper, ConnKeysKCSB
from .kql_response import KqlQueryResponse, KqlError
from .constants import Constants, ConnStrKeys, Cloud
from ._version import __version__
from .log import logger
from .exceptions import KqlEngineError
from .my_utils import json_dumps 
from .kql_client import KqlClient


class AadContext(object):
    """
    Object that contains the Aad parts
    """

    def __init__(self, conn_kv:Dict[str,str], auth_resource:str, client_id:str, options:Dict[str,any]):
        self.kcsb = ConnKeysKCSB(conn_kv, auth_resource)
        self.client_id = client_id
        self.options = options


    def __eq__(self, other:any)-> bool:
        if not isinstance(other, AadContext):
            return False
        if self.client_id != other.client_id:
            return False
        if self.kcsb != other.kcsb:
            return False
        return True


    def __hash__(self):
        return hash((self.client_id, self.kcsb))


class KustoClient(KqlClient):
    """
    Kusto client wrapper for Python."""
 
    _ADX_CLIENT_BY_CLOUD = {
        Cloud.PUBLIC:      "db662dc1-0cfe-4e1c-a843-19a68e65be58",
        Cloud.MOONCAKE:    "db662dc1-0cfe-4e1c-a843-19a68e65be58",
        Cloud.FAIRFAX:     "730ea9e6-1e1d-480c-9df6-0bb9a90e1a0f",
        Cloud.BLACKFOREST: "db662dc1-0cfe-4e1c-a843-19a68e65be58",
        Cloud.PPE:         "db662dc1-0cfe-4e1c-a843-19a68e65be58",
    }
    _ADX_CLIENT_BY_CLOUD[Cloud.CHINA]      = _ADX_CLIENT_BY_CLOUD[Cloud.MOONCAKE]
    _ADX_CLIENT_BY_CLOUD[Cloud.GOVERNMENT] = _ADX_CLIENT_BY_CLOUD[Cloud.FAIRFAX]
    _ADX_CLIENT_BY_CLOUD[Cloud.GERMANY]    = _ADX_CLIENT_BY_CLOUD[Cloud.BLACKFOREST]

    _MGMT_ENDPOINT_VERSION = "v1"
    _QUERY_ENDPOINT_VERSION = "v2"
    _MGMT_ENDPOINT_TEMPLATE = "{0}/{1}/rest/mgmt"
    _QUERY_ENDPOINT_TEMPLATE = "{0}/{1}/rest/query"


    _ADX_PUBLIC_CLOUD_URL_SUFFIX      = ".windows.net"
    _ADX_MOONCAKE_CLOUD_URL_SUFFIX    = ".chinacloudapi.cn"
    _ADX_BLACKFOREST_CLOUD_URL_SUFFIX = ".cloudapi.de"
    _ADX_FAIRFAX_CLOUD_URL_SUFFIX     = ".usgovcloudapi.net"

    _CLOUD_BY_ADX_HOST_SUFFIX = {
        _ADX_PUBLIC_CLOUD_URL_SUFFIX:      Cloud.PUBLIC,
        _ADX_FAIRFAX_CLOUD_URL_SUFFIX:     Cloud.FAIRFAX,
        _ADX_MOONCAKE_CLOUD_URL_SUFFIX:    Cloud.MOONCAKE,
        _ADX_BLACKFOREST_CLOUD_URL_SUFFIX: Cloud.BLACKFOREST
    }


    _ADX_URL_SUFFIX_BY_CLOUD = {
        Cloud.PUBLIC:      _ADX_PUBLIC_CLOUD_URL_SUFFIX,
        Cloud.MOONCAKE:    _ADX_MOONCAKE_CLOUD_URL_SUFFIX,
        Cloud.FAIRFAX:     _ADX_FAIRFAX_CLOUD_URL_SUFFIX,
        Cloud.BLACKFOREST: _ADX_BLACKFOREST_CLOUD_URL_SUFFIX
    }
    _ADX_URL_SUFFIX_BY_CLOUD[Cloud.CHINA]      = _ADX_URL_SUFFIX_BY_CLOUD[Cloud.MOONCAKE]
    _ADX_URL_SUFFIX_BY_CLOUD[Cloud.GOVERNMENT] = _ADX_URL_SUFFIX_BY_CLOUD[Cloud.FAIRFAX]
    _ADX_URL_SUFFIX_BY_CLOUD[Cloud.GERMANY]    = _ADX_URL_SUFFIX_BY_CLOUD[Cloud.BLACKFOREST]


    _DATA_SOURCE_TEMPLATE = "https://{0}.kusto{1}"

    _WEB_CLIENT_VERSION = __version__

    _FQN_DRAFT_PROXY_CLUSTER_PATTERN = re.compile(r"http(s?)\:\/\/ade\.(int\.)?(applicationinsights|loganalytics)\.(?P<host_suffix>(io|cn|us|de)).*$")
    _FQN_DRAFT_PROXY_CLUSTER_PATTERN2 = re.compile(r"http(s?)\:\/\/adx\.(int\.)?monitor\.azure\.(?P<host_suffix>(com|cn|us|de)).*$")

    _CLOUD_BY_ADXPROXY_HOST_SUFFIX = {
        "com": Cloud.PUBLIC,
        "io":  Cloud.PUBLIC,
        "us":  Cloud.FAIRFAX,
        "cn":  Cloud.MOONCAKE,
        "de":  Cloud.BLACKFOREST
    }

    _aad_helpers:Dict[AadContext,_MyAadHelper] = dict()


    def __init__(self, cluster_name:str, conn_kv:Dict[str,str], **options)->None:
        """
        Kusto Client constructor.

        Parameters
        ----------
        kusto_cluster : str
            Kusto cluster endpoint. Example: https://help.kusto.windows.net
        client_id : str
            The AAD application ID of the application making the request to Kusto
        client_secret : str
            The AAD application key of the application making the request to Kusto.
            if this is given, then username/password should not be.
        username : str
            The username of the user making the request to Kusto.
            if this is given, then password must follow and the client_secret should not be given.
        password : str
            The password matching the username of the user making the request to Kusto
        authority : 'microsoft.com', optional
            In case your tenant is not microsoft please use this param.
        """

        super(KustoClient, self).__init__()
        self.default_cloud = options.get("cloud")
        cluster_name = cluster_name or conn_kv[ConnStrKeys.CLUSTER]

        if cluster_name.find("://") > 0:
            data_source = cluster_name
        elif cluster_name.find(".kusto.") > 0:
            data_source = f"https://{cluster_name}"
        elif cluster_name.find(".kusto(mfa).") > 0:
            data_source = f"https://{cluster_name}"
        elif cluster_name.find(".kustomfa.") > 0:
            data_source = f"https://{cluster_name}"
        else:
            adx_url_suffix = self._ADX_URL_SUFFIX_BY_CLOUD.get(self.default_cloud)
            if not adx_url_suffix:
                raise KqlEngineError(f"adx not supported in cloud {self.default_cloud}")
            if cluster_name.endswith(adx_url_suffix):
                data_source = f"https://{cluster_name}"
            else:
                data_source = self._DATA_SOURCE_TEMPLATE.format(cluster_name, adx_url_suffix)

        self._mgmt_endpoint = self._MGMT_ENDPOINT_TEMPLATE.format(data_source, self._MGMT_ENDPOINT_VERSION)
        self._query_endpoint = self._QUERY_ENDPOINT_TEMPLATE.format(data_source, self._QUERY_ENDPOINT_VERSION)

        match = self._FQN_DRAFT_PROXY_CLUSTER_PATTERN.match(data_source) or self._FQN_DRAFT_PROXY_CLUSTER_PATTERN2.match(data_source)
        if match:
            cloud = self._CLOUD_BY_ADXPROXY_HOST_SUFFIX.get(match.group("host_suffix")) or self.default_cloud
            cloud_url_suffix = self._ADX_URL_SUFFIX_BY_CLOUD.get(cloud)
            auth_resource = f"https://kusto.kusto{cloud_url_suffix}"
        else:
            auth_resource = data_source
        
        cloud = self.getCloudFromHost(auth_resource)
        client_id = self._ADX_CLIENT_BY_CLOUD[cloud]
        self._aad_helper = self._get_aad_helper(conn_kv, auth_resource, client_id, **options) if conn_kv.get(ConnStrKeys.ANONYMOUS) is None else None
        self._data_source = data_source


    def _get_aad_helper(self, conn_kv:Dict[str,str], new_auth_resource:str, client_id:str, **options):
        for aad_context, aad_helper in self._aad_helpers.items():
            aad_auth_resource = aad_context.kcsb.data_source
            auth_resource = self._get_auth_resource(new_auth_resource, aad_auth_resource)
            new_context = AadContext(conn_kv, auth_resource, client_id, options)
            if (new_context == aad_context):
                return aad_helper
        
        new_context = AadContext(conn_kv, new_auth_resource, client_id, options)
        http_client = self._http_client if options.get("auth_use_http_client") else None
        new_aad_helper = _MyAadHelper(new_context.kcsb, client_id, http_client=http_client, **options)
        self._aad_helpers[new_context] = new_aad_helper
        return new_aad_helper


    def _get_auth_resource(self, new_auth_resource:str, aad_auth_resource:str)->str:
        aad_parts = aad_auth_resource.split(".")
        new_parts = new_auth_resource.split(".")
        if len(aad_parts) == len(new_parts):
            aad_suffix = ".".join(aad_parts[1:])
            new_suffix = ".".join(new_parts[1:])
            if aad_suffix == new_suffix:
                return aad_auth_resource

        return new_auth_resource


    @property
    def data_source(self)->str:
        return self._data_source


    @property 
    def deep_link_data_source(self)->str:
        match = self._FQN_DRAFT_PROXY_CLUSTER_PATTERN.match(self.data_source) or self._FQN_DRAFT_PROXY_CLUSTER_PATTERN2.match(self.data_source)
        if match:
            cloud = self._CLOUD_BY_ADXPROXY_HOST_SUFFIX.get(match.group("host_suffix")) or self.default_cloud
            cloud_url_suffix = self._ADX_URL_SUFFIX_BY_CLOUD.get(cloud)
            return f"https://help.kusto{cloud_url_suffix}"
        else:
            return self._data_source


    def getCloudFromHost(self, host:str)->str:
        for adx_host_suffix in self._CLOUD_BY_ADX_HOST_SUFFIX:
            if host.endswith(adx_host_suffix):
                return self._CLOUD_BY_ADX_HOST_SUFFIX[adx_host_suffix]
        return Cloud.PUBLIC


    def execute(self, kusto_database:str, kusto_query:str, accept_partial_results:bool=False, **options)->KqlQueryResponse:
        """ 
        Execute a simple query or management command

        Parameters
        ----------
        kusto_database : str
            Database against query will be executed.
        query : str
            Query to be executed
        accept_partial_results : bool
            Optional parameter. If query fails, but we receive some results, we consider results as partial.
            If this is True, results are returned to client, even if there are exceptions.
            If this is False, exception is raised. Default is False.
        options["timeout"] : float, optional
            Optional parameter. Network timeout in seconds. Default is no timeout.
        """
        if kusto_query.startswith("."):
            endpoint_version = self._MGMT_ENDPOINT_VERSION
            endpoint = self._mgmt_endpoint  
        else:
            endpoint_version = self._QUERY_ENDPOINT_VERSION
            endpoint = self._query_endpoint

        # print("### db: ", kusto_database, " ###")
        # print("### csl: ", kusto_query, " ###")
        # kusto_database = kusto_database.replace(" ", "")
        # print("### db: ", kusto_database, " ###")

        request_payload = {
            "db": kusto_database, 
            "csl": kusto_query,
        }

        client_version = f"{Constants.MAGIC_CLASS_NAME}.Python.Client:{self._WEB_CLIENT_VERSION}"

        client_request_id = f"{Constants.MAGIC_CLASS_NAME}.execute"
        client_request_id_tag = options.get("request_id_tag")
        if client_request_id_tag is not None:
            client_request_id = f"{client_request_id};{client_request_id_tag};{str(uuid.uuid4())}/{self._session_guid}/AzureDataExplorer"
        else:
            client_request_id = f"{client_request_id};{str(uuid.uuid4())}/{self._session_guid}/AzureDataExplorer"
            
        app = f'{Constants.MAGIC_CLASS_NAME};{options.get("notebook_app")}'
        app_tag = options.get("request_app_tag")
        if app_tag is not None:
            app = f"{app};{app_tag}"
            
        user_agent = f'{Constants.MAGIC_CLASS_NAME}/{self._WEB_CLIENT_VERSION}'
        user_agent_tag = options.get("request_user_agent_tag")
        if user_agent_tag is not None:
            user_agent = user_agent_tag

        query_properties:dict = options.get("query_properties")  or {}

        if type(kusto_query) == str:
            first_word = kusto_query.split(maxsplit=1)[0].upper()
            # ADX SQL mode
            if first_word in ["SELECT", "UPDATE", "CREATE", "DELETE", "EXPLAIN"]:
                # SQL to Kusto cheat sheet: https://docs.microsoft.com/en-us/azure/data-explorer/kusto/query/sqlcheatsheet
                # MS-TDS/T-SQL Differences between Kusto Microsoft SQL Server: https://docs.microsoft.com/en-us/azure/data-explorer/kusto/api/tds/sqlknownissues
                query_properties["query_language"] = "sql"

        cache_max_age = options.get("request_cache_max_age")
        if cache_max_age is not None and cache_max_age > 0:
            query_properties["query_results_cache_max_age"] = query_properties.get("query_results_cache_max_age")\
                                                              or f"{cache_max_age}s"

        if len(query_properties) > 0:
            properties = {
                "Options": query_properties,
                "Parameters": {},
                "ClientRequestId": client_request_id
            }
            request_payload["properties"] = json_dumps(properties)

        request_headers = {
            "User-Agent": user_agent,
            # "Accept": "application/json",
            # "Accept-Encoding": "gzip,deflate",
            # "Content-Type": "application/json; charset=utf-8",
            "x-ms-client-version": client_version,
            "x-ms-client-request-id": client_request_id,
            "x-ms-app": app
        }
        user_tag = options.get("request_user_tag")
        if user_tag is not None:
            request_headers["x-ms-user"] = user_tag
        if self._aad_helper is not None:
            request_headers["Authorization"] = self._aad_helper.acquire_token()
            request_headers["Fed"] = "True"

        cache_max_age = options.get("request_cache_max_age")
        if cache_max_age is not None:
            if cache_max_age > 0:
                request_headers["Cache-Control"] = f"max-age={cache_max_age}"
            else:
                request_headers["Cache-Control"] = "no-cache"
            
        # print("endpoint: ", endpoint)
        # print("headers: ", request_headers)
        # print("payload: ", request_payload)
        # print("timeout: ", options.get("timeout"))

        log_request_headers = request_headers
        if request_headers.get("Authorization"):
            log_request_headers = request_headers.copy()
            log_request_headers["Authorization"] = "..."  

        logger().debug(f"KustoClient::execute - POST request - url: {endpoint}, headers: {log_request_headers}, payload: {request_payload}, timeout: {options.get('timeout')}")

        # collect this information, in case bug report will be generated
        KqlClient.last_query_info = {
            "request": {
                "endpoint": endpoint,
                "headers": log_request_headers,
                "payload": request_payload,
                "timeout": options.get("timeout"),
            }
        }

        response = self._http_client.post(endpoint, headers=request_headers, json=request_payload, timeout=options.get("timeout"))

        logger().debug(f"KustoClient::execute - response - status: {response.status_code}, headers: {response.headers}, payload: {response.text}")

        # print("response status code: ", response.status_code)
        # print("response", response)
        # print("response text", response.text)

        # collect this information, in case bug report will be generated
        self.last_query_info["response"] = {  # pylint: disable=unsupported-assignment-operation
            "status_code": response.status_code
        }

        if response.status_code < 200  or response.status_code >= 300:  # pylint: disable=E1101
            try:
                parsed_error = json.loads(response.text)
            except:
                parsed_error = response.text
            # collect this information, in case bug report will be generated
            self.last_query_info["response"]["error"] = parsed_error  # pylint: disable=unsupported-assignment-operation, unsubscriptable-object
            raise KqlError(response.text, response)

        kql_response = KqlQueryResponse(response.json(), endpoint_version)

        if kql_response.has_exceptions() and not accept_partial_results:
            try:
                error_message = json_dumps(kql_response.get_exceptions())
            except:
                error_message = str(kql_response.get_exceptions())
            raise KqlError(error_message, response, kql_response)

        return kql_response



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

from typing import Any, Dict, List, Type
import warnings
import re
import platform


from ._debug_utils import debug_print
from ._require import EXTRAS_REQUIRE, INSTALL_REQUIRES, list_union, strip_package_name
from .constants import Constants
from .log import logger, isNullLogger
from .my_utils import get_env_var_list, get_env_var_bool


saved_formatwarning = warnings.formatwarning



class Dependencies(object):

    IGNORE_WARNINGS = "ignore"
    SHOW_WARNINGS = "default"
    RAISE_WARNINGS = "error"

    PREFIX_TAG = "###"
    MANDATORY_TAG = "#M#"  # Kqlmagic cannot work without it
    BASE_TAG = "#B#"       # Kqlmagic base feature, however they may be disabled
    OPTIONAL_TAG = "#O#"   # Kqlmagic optional feature
    EXTRA_TAG = "#E#"
    DISABLED_TAG = "#D#"

    MANDATORY_MODULES_REGEX = r'^' + MANDATORY_TAG
    BASE_MODULES_REGEX = r'^' + BASE_TAG
    OPTIONAL_MODULES_REGEX = r'^' + OPTIONAL_TAG
    EXTRA_MODULES_REGEX = r'^' + EXTRA_TAG
    DISABLED_MODULES_REGEX = r'^' + DISABLED_TAG
    ALL_MODULES_REGEX = r'^(' + f'{MANDATORY_TAG}|{BASE_TAG}|{OPTIONAL_TAG}|{EXTRA_TAG}|{DISABLED_TAG}' + r')'

    MISSING_ENV_TAG = "#V#"
    MISSING_ENV_REGEX = r'^' + MISSING_ENV_TAG

    BAD_ENV_TAG = "#A#"
    BAD_ENV_REGEX = r'^' + BAD_ENV_TAG


    VERSION_IN_MODULE=True

    # platforms: 'Linux' | 'Windows' | 'Darwin' | 'Java' | ''
    platform_dependencies = {
        "Windows": (
            ('msal_extensions.windows', 'msal_extensions', OPTIONAL_TAG, "won't be able to authenticate using msal Single-Sign-On authentication modes", 'msal_extensions'),
            # ('winwin', 'winwin', OPTIONAL_TAG, "will be dsiabled", VERSION_IN_MODULE),
        ),
        "Linux": (
            ('msal_extensions.libsecret', 'msal_extensions', OPTIONAL_TAG, "won't be able to authenticate using msal Single-Sign-On authentication modes", 'msal_extensions'),
            # ('winwin', 'winwin', OPTIONAL_TAG, "will be dsiabled", VERSION_IN_MODULE),
        ),
        "Darwin": (
            ('msal_extensions.osx', 'msal_extensions', OPTIONAL_TAG, "won't be able to authenticate using msal Single-Sign-On authentication modes", 'msal_extensions'),
            # ('winwin', 'winwin', OPTIONAL_TAG, "will be dsiabled", VERSION_IN_MODULE),
        )
    }

    dependencies: list = [
        ('dateutil.parser', 'python-dateutil', MANDATORY_TAG, "won't be able handle datetime properly", 'dateutil'),
        ('traitlets', 'traitlets', MANDATORY_TAG, "won't be able to use execute Kqlmagic", VERSION_IN_MODULE),

        ('prettytable', 'prettytable', OPTIONAL_TAG, "won't be able to display tables", VERSION_IN_MODULE),
        # ('tabulate', 'tabulate', OPTIONAL_TAG, "won't be able to display tables", VERSION_IN_MODULE),
        ('requests', 'requests', OPTIONAL_TAG, "will use urllib, that might have ssl and or proxies restrictions", VERSION_IN_MODULE),
        ('msal', 'msal', OPTIONAL_TAG, "won't be able to authenticate using msal authentication modes, and Kqlmagic sso will be disabled", VERSION_IN_MODULE),
        ('azure.identity', 'azure-identity', OPTIONAL_TAG, "Some authentication options won't be available", VERSION_IN_MODULE),
        ('pandas', 'pandas', OPTIONAL_TAG, "won't be able to use dataframes", VERSION_IN_MODULE),
        ('IPython', 'ipython', OPTIONAL_TAG, "won't be to execute as an jupyter magic", VERSION_IN_MODULE),
        ('ipykernel', 'ipykernel', OPTIONAL_TAG, "won't be to execute as an jupyter magic on some jupyter variants", VERSION_IN_MODULE),
        ('pygments', 'pygments', OPTIONAL_TAG, "json objects won't be decorated with colors", VERSION_IN_MODULE),
        ('pygments.lexers.data', 'pygments', OPTIONAL_TAG, "json objects won't be decorated with colors", 'pygments'),
        ('pygments.formatters.terminal', 'pygments', OPTIONAL_TAG, "json objects won't be decorated with colors", 'pygments'),
        ('pyperclip', 'pyperclip', OPTIONAL_TAG, "copy/paste feature will be disabled in device code authentication", VERSION_IN_MODULE),
        ('azure.common.credentials', 'azure-common', OPTIONAL_TAG, "-try_azcli_login, -try_azcli_login_subscription and -try_azcli_login_by_profile authentication options will be dsiabled", 'azure.common'),
        ('msrestazure.azure_active_directory', 'msrestazure', OPTIONAL_TAG, "-try_msi authentication options will be dsiabled", 'msrestazure'),
        ('psutil', 'psutil', OPTIONAL_TAG, "some jupyter variants may not be detected correctly", VERSION_IN_MODULE),

        ('matplotlib.pyplot', 'matplotlib', DISABLED_TAG, 'plotting with matplotlib will be dsiabled', 'matplotlib'),

        ('matplotlib.cm', 'matplotlib', OPTIONAL_TAG, "matplotlib color maps (palettes) won't be available in plots", 'matplotlib'),
        ('matplotlib.colors', 'matplotlib', OPTIONAL_TAG, "matplotlib color maps (palettes) won't be available in plots", 'matplotlib'),
        ('pkg_resources', 'setuptools', OPTIONAL_TAG, 'may not detect new version exist in PyPI', 'setuptools'),
        ('plotly', 'plotly', OPTIONAL_TAG, "won't display charts with plotly", VERSION_IN_MODULE),
        ('plotly.graph_objs', 'plotly', OPTIONAL_TAG, "won't display charts with plotly", 'plotly'),
        ('flask', 'flask', OPTIONAL_TAG, "popups and some authentication modes won't work on some local jupyter variants", VERSION_IN_MODULE),
        ('isodate', 'isodate', OPTIONAL_TAG, "Azure Monitor AI/LA timespan specified with python timedate.timedelta object will be disabled", VERSION_IN_MODULE),
        ('markdown', 'markdown', OPTIONAL_TAG, "in text jupyter implementations, some help information might not be nicely displayed", VERSION_IN_MODULE),
        ('bs4', 'beautifulsoup4', OPTIONAL_TAG, "in text jupyter implementations, some help information might not be nicely displayed", VERSION_IN_MODULE),
        ('lxml', 'lxml', OPTIONAL_TAG, "in text jupyter implementations, some help information won't be nicely displayed", VERSION_IN_MODULE),
        ('ipywidgets', 'ipywidgets', OPTIONAL_TAG, "widget features will be disabled", VERSION_IN_MODULE),
        ('cryptography.fernet', 'cryptography', OPTIONAL_TAG, "Single Sign On feature will be disabled", 'cryptography'),
        ('cryptography.hazmat.backends', 'cryptography', OPTIONAL_TAG, "Single Sign On feature will be disabled", 'cryptography'),
        ('cryptography.hazmat.primitives', 'cryptography', OPTIONAL_TAG, "Single Sign On feature will be disabled", 'cryptography'),
        ('cryptography.hazmat.primitives.kdf.pbkdf2', 'cryptography', OPTIONAL_TAG, "Single Sign On feature will be disabled", 'cryptography'),
        ('password_strength', 'password_strength', OPTIONAL_TAG, "Single Sign On feature will be disabled", VERSION_IN_MODULE),
    ]

    dependencies_dict = {entry[0]:entry for entry in dependencies}


    installed_modules:Dict[str,Any] = {}
    installed_versions:Dict[str,str] = {}

    extras_names:List[str] = []
    extras_require_packages:List[str] = []
    extras_require_package_names:List[str] = []

    install_requires_packages:List[str] = []
    install_requires_package_names:List[str] = []

    install_package_names:List[str] = []

    is_only_installed_packages:bool = False
    debug_disabled_packages:List[str] = []


    def __init__(self)->None:
        self.extend_dependencies()
        self.set_installed_modules_and_versions()
        # Dependencies.installed_modules["pandas"] = False
        # Dependencies.installed_modules["plotly"] = False


    @classmethod
    def extend_dependencies(cls)->None:
        key = platform.system()
        cls.dependencies.extend(cls.platform_dependencies.get(key, ()))


    @classmethod
    def set_installed_modules_and_versions(cls)->None:
        cls.debug_disabled_packages = [name.strip() for name in (get_env_var_list(f"{Constants.MAGIC_CLASS_NAME_UPPER}_DEBUG_DISABLE_PACKAGES") or [])]
        cls.is_only_installed_packages = get_env_var_bool(f"{Constants.MAGIC_CLASS_NAME_UPPER}_ONLY_INSTALLED_PACKAGES", False)

        cls.extras_names = ([name.strip() for name in (get_env_var_list(f"{Constants.MAGIC_CLASS_NAME_UPPER}_EXTRAS_REQUIRE") or [])] or
                            [name.strip() for name in (get_env_var_list(f"{Constants.MAGIC_CLASS_NAME_UPPER}_EXTRAS_REQUIRES") or [])])

        if cls.extras_names:
            cls.extras_require_packages = list_union(*[EXTRAS_REQUIRE.get(name, []) for name in cls.extras_names])
        else:
            python_branch = platform.python_branch()
            name = 'saw' if python_branch.endswith(Constants.SAW_PYTHON_BRANCH_SUFFIX) else 'default'
            cls.extras_require_packages = EXTRAS_REQUIRE.get(name, [])
        cls.extras_require_package_names = list(map(strip_package_name, cls.extras_require_packages))

        cls.install_requires_packages = INSTALL_REQUIRES
        cls.install_requires_package_names = list(map(strip_package_name, cls.install_requires_packages))

        cls.install_package_names = list_union(cls.install_requires_package_names, cls.extras_require_package_names)

        for item in cls.dependencies:
            module_name = item[0]
            package_name = item[1]
            version_location = item[4]
            cls.get_module(module_name, package_name=package_name, version_location=version_location, dont_throw=True)
        logger().debug(f"installed versions: {cls.installed_versions}")


    @classmethod
    def is_installed(cls, module_name:str)->bool:
        return cls.get_module(module_name, dont_throw=True) is not None


    @classmethod
    def installed_packages(cls)->Dict[str,str]:
        return {package: cls.installed_versions.get(package) for package in cls.installed_versions}



    @classmethod
    def get_module(cls, module_name:str, package_name:str=None, version_location:str=None, dont_throw:bool=False, message:str=None):
        try:
            module = cls.installed_modules.get(module_name)
            if module:
                return module

            elif module is False:
                raise Exception("not_installed")

            if cls.debug_disabled_packages and (package_name.replace("_", "-") in cls.debug_disabled_packages or package_name.replace("-", "_") in cls.debug_disabled_packages):
                raise Exception("debug_disabled_package")
                
            if cls.is_only_installed_packages and package_name.replace("_", "-") not in  cls.install_package_names and package_name.replace("-", "_") not in  cls.install_package_names:
                debug_print(f">>> package {package_name} disabled by due to only install packages set to True")
                raise Exception(f"package {package_name} disabled by due to only install packages set to True")

            import importlib
            module = importlib.import_module(module_name)
            cls.installed_modules[module_name] = module
            if package_name is not None and cls.installed_versions.get(package_name) is None:
                version = None
                if version_location is not None:
                    try:
                        if version_location == cls.VERSION_IN_MODULE:
                            version_module = module
                        else:
                            # version_location should refer to a module 
                            version_module = cls.installed_modules.get(version_location)
                            if version_module is None:
                                version_module = importlib.import_module(version_location)
                                cls.installed_modules[version_location] = version_module
                        version = version_module.__version__
                    except:
                        try:
                            import pkg_resources  # part of setuptools
                            version = pkg_resources.require(package_name)[0].version
                        except:
                            pass
                cls.installed_versions[package_name] = version or "?.?.?"
            return module
        except Exception as error:
            cls.installed_modules[module_name] = False
            if f"{error}" == "debug_disabled_package":
                debug_disabled_package_message = f"package '{package_name}' was disabled, because specified in enironment variable {Constants.MAGIC_CLASS_NAME_UPPER}__DEBUG_DISABLE_PACKAGES"
                print(f"!!! Note: {debug_disabled_package_message} !!!")
                message = f", {message}" if message else ""
                message = f"{debug_disabled_package_message}{message}"
            if dont_throw:
                return None
            item = cls.dependencies_dict.get(module_name)
            message = f", {message}" if message else ""
            import_error_message = cls._import_error_message(*item) if item else f"due to '{module_name}' module/package not installed"
            import_error_message = f"{import_error_message}{message}"
            raise NotImplementedError(import_error_message)


    @classmethod
    def warn_missing_dependencies(cls, options:Dict[str,Any]=None)->None:
        options = options or {}
        if not isNullLogger:
            for item in cls.dependencies:
                module_name = item[0]
                if cls.is_installed(module_name):
                    package_name = item[1]
                    tail = f"installed version '{cls.installed_versions.get(package_name)}'"
                else:
                    tail = "NOT installed"
                logger().debug(f"dependency -- '{item}' -- {tail}")

        warnings.filterwarnings(cls.RAISE_WARNINGS, message=cls.MANDATORY_MODULES_REGEX, category=ImportWarning)

        warnings.filterwarnings(cls.SHOW_WARNINGS, message=cls.BASE_MODULES_REGEX, category=ImportWarning)

        warnings.filterwarnings(cls.SHOW_WARNINGS, message=cls.OPTIONAL_MODULES_REGEX, category=ImportWarning)

        warnings.filterwarnings(cls.IGNORE_WARNINGS, message=cls.EXTRA_MODULES_REGEX, category=ImportWarning)

        warnings.filterwarnings(cls.IGNORE_WARNINGS, message=cls.DISABLED_MODULES_REGEX, category=ImportWarning)
        # warnings.simplefilter("always")
        warnings.formatwarning = cls._formatwarning

        if cls.extras_names:
            cls.warn_unknown_extras_require_name(cls.extras_names)

        for item in cls.dependencies:
            cls._import_warn(cls.install_package_names, *item)


    @classmethod
    def _import_error_message(cls, module_name:str, package:str, tag:str, message:str, version_location:str)->str:
        return f"failed to import '{module_name}' from '{package}', {message}"


    @classmethod
    def _import_warn(cls, install_package_names:List[str], module_name:str, package:str, tag:str, message:str, version_location:str)->None:
        if not cls.installed_modules.get(module_name) and package in install_package_names:
            import_error_message = cls._import_error_message(module_name, package, tag, message, version_location)
            warnings.warn_explicit(f"{tag}{import_error_message}", ImportWarning, '', 0)


    @classmethod
    def _formatwarning(cls, message:str, category:Type[Warning], filename:str, lineno:int, line:str=None)->str:
        global saved_formatwarning
        if category.__name__ == 'ImportWarning' and re.match(cls.ALL_MODULES_REGEX, f'{message}'):
            message = f'{message}'[3:]
            return f"{category.__name__}: {message}\n"
        elif category.__name__ == 'UserWarning' and re.match(cls.MISSING_ENV_REGEX, f'{message}'):
            message = f'{message}'[3:]
            # assumes env variable is passed as filename
            return f"MissingEnvironmentVariable: {filename} is not set. {message}\n"
        elif category.__name__ == 'UserWarning' and re.match(cls.BAD_ENV_REGEX, f'{message}'):
            message = f'{message}'[3:]
            # assumes env variable is passed as filename
            return f"BadEnvironmentVariable: {filename} value is invalid. {message}\n"
        else:
            return saved_formatwarning(message, category, filename, lineno, line=line)


    @classmethod
    def warn_missing_env_variables(cls, options:Dict[str,Any]=None)->None:
        options = options or {}
        if options.get("notebook_app") in ["azureml", "azuremljupyternotebook", "azuremljupyterlab"]:
            if options.get("notebook_service_address") is None:
                var_name = f"{Constants.MAGIC_CLASS_NAME_UPPER}_AZUREML_COMPUTE"
                message = f"{cls.MISSING_ENV_TAG}Popup windows might not work properly (copy/paste compute host value from the address bar of the notebook when activated in azureml Jupyter mode)"
                warnings.warn_explicit(f"{message}", UserWarning, var_name, 0)

    @classmethod
    def warn_unknown_extras_require_name(cls, extras_names:List[str])->None:
        unknown_names = [name for name in extras_names if name not in EXTRAS_REQUIRE]
        if unknown_names:
            var_name = f"{Constants.MAGIC_CLASS_NAME_UPPER}_EXTRAS_REQUIRE"
            message = f"{cls.BAD_ENV_TAG}unknown extras_requires {unknown_names}, wrong warning about missing modules might display"
            warnings.warn_explicit(f"{message}", UserWarning, var_name, 0)




# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

# DPAPI access library
# This file uses code originally created by Crusher Joe:
# http://article.gmane.org/gmane.comp.python.ctypes/420
#

import platform
import getpass


if platform.system() == 'Windows':
    try:
        from ctypes import Structure, POINTER, byref, create_string_buffer, windll, cdll
        from ctypes.wintypes import DWORD, PCHAR

        LocalFree = windll.kernel32.LocalFree
        memcpy = cdll.msvcrt.memcpy
        CryptProtectData = windll.crypt32.CryptProtectData
        CryptUnprotectData = windll.crypt32.CryptUnprotectData
    except Exception:
        dpapi_installed = False
    else:
        dpapi_installed = True
else:
    dpapi_installed = False


from .log import logger
from .constants import DpapiParam    


if dpapi_installed:

    CRYPTPROTECT_UI_FORBIDDEN = 0x01

    class DATA_BLOB(Structure):  # type: ignore  reportUnboundVariable

        _fields_ = [("cbData", DWORD), ("pbData", PCHAR)]  # type: ignore  reportUnboundVariable


    class DpapiCrypto(object):

        def __init__(self, **options):
            entropy: str = options.get(DpapiParam.ENTROPY) or "kqlmagic"
            self._entropy_blob_byref = byref(self._toBlob(entropy)) if entropy else None
            self._description: str = options.get(DpapiParam.DESCRIPTION)


        @property
        def suffix(self) -> str:
            return getpass.getuser()


        def _getData(self, blob)-> bytes:
            blob_length = int(blob.cbData)
            pbData = blob.pbData
            buffer = create_string_buffer(blob_length)
            memcpy(buffer, pbData, blob_length)
            LocalFree(pbData)
            return buffer.raw


        def _toBlob(self, data):
            data_bytes = str(data).encode() if isinstance(data, str) else data
            data_buffer = create_string_buffer(data_bytes)
            return DATA_BLOB(len(data_bytes), data_buffer)


        def encrypt(self, data: str)-> bytes:
            logger().debug(f"DpapiCrypto::encrypt(self, data)")
            if data:
                data_blob = self._toBlob(data)
                encrypted_blob = DATA_BLOB()
                if CryptProtectData(byref(data_blob), self._description, self._entropy_blob_byref, None, None,
                                    CRYPTPROTECT_UI_FORBIDDEN, byref(encrypted_blob)):
                    return self._getData(encrypted_blob)


        def decrypt(self, encrypted_data_bytes: bytes)-> str:
            logger().debug(f"DpapiCrypto::decrypt(self, encrypted_data_bytes)")
            if encrypted_data_bytes:
                encrypted_data_blob = self._toBlob(encrypted_data_bytes)
                data_blob = DATA_BLOB()
                if CryptUnprotectData(byref(encrypted_data_blob), None, self._entropy_blob_byref, None, None,
                                      CRYPTPROTECT_UI_FORBIDDEN, byref(data_blob)):
                    data_bytes = self._getData(data_blob)
                    return data_bytes.decode()


        def verify(self, encrypted_data_bytes: bytes) -> None:
            pass

else:
    class DpapiCrypto(object):

        def __init__(self, **options):
            pass



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

from typing import Any, List, Dict, Union, Tuple
import re
import itertools
import getpass


from .engine import Engine
from .kql_response import KqlQueryResponse, KqlSchemaResponse
from .kql_proxy import KqlResponse
from .kql_client import KqlClient
from .constants import ConnStrKeys, Schema
from .exceptions import KqlEngineError
from .my_utils import get_valid_name, adjust_path
from .parser import Parser
from .log import logger


_FQN_KUSTO_CLUSTER_PATTERN        = re.compile(r"(http(s?)\:\/\/)?(?P<cname>.*)\.kusto\.(windows\.net|chinacloudapi.cn|cloudapi.de|usgovcloudapi.net)$")
_FQN_DRAFT_PROXY_CLUSTER_PATTERN  = re.compile(r"http(s?)\:\/\/ade\.(int\.)?(?P<io>(applicationinsights|loganalytics))\.io\/subscriptions\/(?P<subscription>.*)$")
_FQN_DRAFT_PROXY_CLUSTER_PATTERN2 = re.compile(r"http(s?)\:\/\/adx\.(int\.)?(?P<io>(monitor))\.azure\.(com|us|cn)\/subscriptions\/(?P<subscription>.*)$")
_FQN_ARIA_KUSTO_CLUSTER_PATTERN   = re.compile(r"http(s?)\:\/\/kusto\.aria\.microsoft\.com$")


class KqlEngine(Engine):

    _ALT_URI_SCHEMA_NAMES:List[str] = []
    _RESERVED_CLUSTER_NAME:str = None
    _URI_SCHEMA_NAME:str = None
    _MANDATORY_KEY:str = None
    _VALID_KEYS_COMBINATIONS:List[List[str]] = []


    # Object constructor
    def __init__(self):
        super(KqlEngine, self).__init__()
        self.bind_url = None
        self._parsed_conn = {}
        self._keys_set = set()
        self.database_name = None
        self.client_database_name = None
        self.database_friendly_name = None
        self.cluster_name = None
        self.cluster_friendly_name = None
        self.alias = None
        self.client = None
        self.options = {}

        self.validated = None
        self.conn_name = None
        self.id = None


    @classmethod
    def get_alt_uri_schema_names(cls)->List[str]:
        return cls._ALT_URI_SCHEMA_NAMES 


    @classmethod
    def get_reserved_cluster_name(cls)->str:
        return cls._RESERVED_CLUSTER_NAME


    @classmethod
    def get_mandatory_key(cls)->str:
        return cls._MANDATORY_KEY


    @classmethod
    def get_valid_keys_combinations(cls)->List[List[str]]:
        return cls._VALID_KEYS_COMBINATIONS


    @classmethod
    def get_uri_schema_name(cls):
        return cls._URI_SCHEMA_NAME


    def obfuscate_parsed_conn(self)->Dict[str,str]:
        parse_conn_copy = {**self._parsed_conn}
        for p in parse_conn_copy:
            if p in [ConnStrKeys.PASSWORD, ConnStrKeys.CLIENTSECRET, ConnStrKeys.CERTIFICATE, ConnStrKeys.APPKEY]:
                parse_conn_copy[p] = "*****"
        return parse_conn_copy


    # collect datails, in case bug report will be generated
    def get_details(self)->Dict[str,Any]:
        details =  {
            "cluster_name": self.cluster_name,
            "database_name": self.database_name,
            "parsed_conn": self.obfuscate_parsed_conn(),
        }
        client = self.get_client()
        if client and client._aad_helper:
            details["auth"] = client._aad_helper.get_details()
        return details


    def __eq__(self, other)->bool:
        return self.get_id() == other.get_id()


    def get_alias(self)->str:
        return self.alias


    def get_database_name(self)->str:
        if not self.database_name:
            raise KqlEngineError("Database is not defined.")
        return self.database_name


    def get_client_database_name(self)->str:
        name = self.client_database_name or self.database_name
        if name is None:
            raise KqlEngineError("Database is not defined.")
        return name


    def get_cluster_name(self)->str:
        if not self.cluster_name:
            raise KqlEngineError("Cluster is not defined.")
        return self.cluster_name


    def get_cluster_friendly_name(self)->str:
        if not self.cluster_friendly_name:
            raise KqlEngineError("Cluster friendly name is not defined.")
        return self.cluster_friendly_name


    def get_database_friendly_name(self)->str:
        if not self.database_friendly_name:
            raise KqlEngineError("Database friendly name is not defined.")
        return self.database_friendly_name


    def get_id(self)->str:
        if self.id is None:
            if self.cluster_friendly_name and self.database_friendly_name:
                self.id = f"{self.database_friendly_name}@{self.cluster_friendly_name}"
            else:
                raise KqlEngineError("Database and/or cluster is not defined.")
        return self.id



    def get_conn_name(self)->str:
        if self.conn_name is None:
            if self.alias and self.cluster_friendly_name and self.database_friendly_name:
                self.conn_name = f"{self.alias}@{self.cluster_friendly_name}"
            else:
                raise KqlEngineError("Database and/or cluster is not defined.")
        return self.conn_name


    def get_deep_link(self, query:str, options:Dict[str,Any]=None)->str:
        options = options or {}
        return None


    def createDatabaseFriendlyName(self, dname:str)->str:
        return get_valid_name(dname)


    def createClusterFriendlyName(self, cname:str)->str:
        name = cname.lower()
        match = _FQN_KUSTO_CLUSTER_PATTERN.match(name)
        if match:
            return match.group("cname")

        name = name[:-1] if name[-1] == "/" else name

        match = _FQN_DRAFT_PROXY_CLUSTER_PATTERN.match(name) or _FQN_DRAFT_PROXY_CLUSTER_PATTERN2.match(name)
        if match:
            components = match.group("subscription").split("/")
            if match.group("io") == "monitor":
                if len(components) >= 5 and components[3] == "providers":
                    resource_name = "app" if components[4] == "microsoft.insights" else "workspace"
                else:
                    resource_name = "apps_or_workspace"
            else:
                resource_name = "app" if match.group("io") == "applicationinsights" else "workspace"
            name = f"{resource_name}s_in_subscription_{components[0]}"
            if len(components) >= 3:
                if components[1] == "resourcegroups":
                    name = f"{name}_resourcegroup_{components[2]}"
                    if len(components) >= 7:
                        name = f"{name}_{resource_name}_{components[6]}"
                else:
                    name = f"{name}_{components[1]}_{components[2]}"
        elif _FQN_ARIA_KUSTO_CLUSTER_PATTERN.match(name):
            name = Schema.ARIA
        else:
            if name.startswith("https://"):
                name = name[8:]
            elif name.startswith("http://"):
                name = name[7:]
            name = name.replace("/", "_").replace(".", "_").replace(":", "_").replace("-", "_")
        return name


    def get_client(self)->KqlClient:
        return self.client


    def client_execute(self, query:str, user_namespace:Dict[str,Any]=None, database:str=None, **options)->Union[KqlQueryResponse, KqlSchemaResponse]:
        if query.strip():
            client = self.get_client()
            if not client:
                raise KqlEngineError("Client is not defined.")
            return client.execute(database or self.get_client_database_name(), query, accept_partial_results=False, **options)


    def execute(self, query:str, user_namespace:Dict[str,Any]=None, database:str=None, **options)->KqlResponse:
        if query.strip():
            response = self.client_execute(query, user_namespace, database, **options)
            return KqlResponse(response, **options)


    def validate(self, **options)->None:
        if not self.options.get("validate_connection_string_done") and options.get("validate_connection_string"):
            query = "range c from 1 to 10 step 1 | count"
            kql_response = self.execute(query, **options)
            table = kql_response.tables[0]
            if table.rowcount() != 1 or table.colcount() != 1 or [r for r in table.fetchall()][0][0] != 10:
                raise KqlEngineError("Client failed to validate connection.")
            self.options["validate_connection_string_done"] = True
        

    def validate_database_name(self, **options)->None:
        if not self.options.get("validate_database_name_done"):
            try:
                database_name_list, databases_by_pretty_name = self._get_databases_by_pretty_name(**options)
                if self.database_name not in database_name_list:
                    if self.database_name in databases_by_pretty_name:
                        self.client_database_name = databases_by_pretty_name[self.database_name]
                    else:
                        raise KqlEngineError("Database wasn't found in cluster.")
            except NotImplementedError:
                pass
            except:
                raise

            self.options["validate_database_name_done"] = True


    def _get_databases_by_pretty_name(self, **options)->Tuple[List[str],Dict[str,str]]:
        raise NotImplementedError(f"_get_databases_by_pretty_name not implemented for this engine.")


    _CREDENTIAL_KEYS = {
        ConnStrKeys.TENANT,
        ConnStrKeys.AAD_URL,
        ConnStrKeys.USERNAME,
        ConnStrKeys.CLIENTID,
        ConnStrKeys.CERTIFICATE,
        ConnStrKeys.CLIENTSECRET,
        ConnStrKeys.APPKEY,
        ConnStrKeys.PASSWORD,
        ConnStrKeys.CERTIFICATE_THUMBPRINT,
    }

    _SECRET_KEYS = {
        ConnStrKeys.CLIENTSECRET, 
        ConnStrKeys.APPKEY, 
        ConnStrKeys.PASSWORD, 
        ConnStrKeys.CERTIFICATE_THUMBPRINT
    }

    _NOT_INHERITABLE_KEYS = {
        ConnStrKeys.APPKEY,
        ConnStrKeys.ALIAS
    }

    _OPTIONAL_KEYS = {
        ConnStrKeys.TENANT, 
        ConnStrKeys.AAD_URL, 
        ConnStrKeys.DATA_SOURCE_URL, 
        ConnStrKeys.ALIAS, 
        ConnStrKeys.CLIENTID
    }

    _INHERITABLE_KEYS = {
        ConnStrKeys.CLUSTER, 
        ConnStrKeys.TENANT,
        ConnStrKeys.AAD_URL,
        ConnStrKeys.DATA_SOURCE_URL
    }

    _EXCLUDE_FROM_URL_KEYS = {
        ConnStrKeys.DATABASE,
        ConnStrKeys.ALIAS
    }
    
    _SHOULD_BE_NULL_KEYS = {
        ConnStrKeys.CODE,
        ConnStrKeys.ANONYMOUS
    }


    def _parse_connection_str(self, conn_str:str, user_ns:Dict[str,Any])->Dict[str,str]:
        # logger().debug(f"kql_engine.py - _parse_connection_str - params:  conn_str: {conn_str}, user_ns: {safe_str(user_ns)}")
        logger().debug(f"kql_engine.py - _parse_connection_str - params:  conn_str: {conn_str}")
        rest = conn_str[conn_str.find("://") + 3:].strip()
        # get key/values in connection string
        parsed_conn_kv = Parser.parse_and_get_kv_string(rest, user_ns)
        logger().debug(f"kql_engine.py - _parse_connection_str - parsed_conn_kv: {parsed_conn_kv} (return of Parser.parse_and_get_kv_string)")
        # In case certificate_pem_file was specified instead of certificate.
        pem_file_name = parsed_conn_kv.get(ConnStrKeys.CERTIFICATE_PEM_FILE)
        if pem_file_name is not None:
            pem_file_name = adjust_path(pem_file_name)
            with open(pem_file_name, "r") as pem_file:
                parsed_conn_kv[ConnStrKeys.CERTIFICATE] = pem_file.read()
                del parsed_conn_kv[ConnStrKeys.CERTIFICATE_PEM_FILE]
        return parsed_conn_kv


    def _check_for_unknown_keys(self, matched_keys_set:set, keys_combinations:List[List[str]])->None:
        # check for unknown keys
        all_keys = set(itertools.chain(*keys_combinations))
        unknown_keys_set = matched_keys_set.difference(all_keys)
        if len(unknown_keys_set) > 0:
            logger().debug(f"kql_engine.py - _check_for_unknown_keys - the following keys are unknow: {unknown_keys_set}")
            raise ValueError(f"detected unknown keys: {unknown_keys_set}")


    def _check_for_mandatory_key(self, matched_keys_set:set, mandatory_key:str)->None:
        # check that mandatory key in matched set
        if mandatory_key not in matched_keys_set:
            logger().debug(f"kql_engine.py - _check_for_mandatory_key - the following mandatory_key is missing: {mandatory_key}")
            raise KqlEngineError(f"mandatory key {mandatory_key} is missing")


    def _find_combination_set(self, current, matched_keys_set:set, parsed_conn_kv:Dict[str,str], keys_combinations:List[List[str]])->set:
        # find a valid combination for the set
        valid_combinations = [c for c in keys_combinations if matched_keys_set.issubset(c)]
        logger().debug(f"kql_engine.py - _find_combination - found these valid_combinations: {valid_combinations}")

        # in case of ambiguity, assume it is based on current connection, resolve by copying missing values from current
        if len(valid_combinations) > 1 and current is not None and matched_keys_set.issubset(current._keys_set):
            failed = False
            saved_parsed_conn_kv = parsed_conn_kv.copy()
            saved_matched_keys_set = matched_keys_set.copy()
            inherited = False
            for k, v in current._parsed_conn.items():
                if k not in matched_keys_set and k not in self._NOT_INHERITABLE_KEYS:
                    parsed_conn_kv[k] = v
                    matched_keys_set.add(k)
                    inherited = True
                    
            if inherited:
                for k in self._CREDENTIAL_KEYS.intersection(matched_keys_set):
                    if parsed_conn_kv[k] != current._parsed_conn.get(k):
                        failed = True
                        # raise KqlEngineError("missing keys.")

                if failed:
                    parsed_conn_kv = saved_parsed_conn_kv
                    matched_keys_set = saved_matched_keys_set
            if not failed:
                logger().debug(f"kql_engine.py - _find_combination - chose current combination_keys_list is: {current._keys_set}")
                return current._keys_set

        # only one combination can be accepted
        if len(valid_combinations) == 0:
            logger().debug(f"kql_engine.py - _find_combination - valid_combinations is empty: {valid_combinations}")
            raise KqlEngineError("not a valid keys set, missing keys")

        combination_keys_list = None
        # if still too many choose the shortest
        if len(valid_combinations) > 1:
            combination_keys_list =  min(valid_combinations, key=len)
        else:
            combination_keys_list = valid_combinations[0]
        logger().debug(f"kql_engine.py - _find_combination - chose combination_keys_list - shortest out of valid_combinations: {combination_keys_list}")

        if combination_keys_list is None:
            raise KqlEngineError("not a valid keys set, missing keys")

        return set(combination_keys_list)


    def _inherit_keys(self, current, matched_keys_set:set, parsed_conn_kv:Dict[str,str])->None:
        # in case inheritable fields are missing inherit from current if exist
        inherit_keys_set = self._INHERITABLE_KEYS.intersection(self._keys_set).difference(matched_keys_set)
        if len(inherit_keys_set) > 1:
            if current is not None:
                for k in inherit_keys_set:
                    v = current._parsed_conn.get(k)
                    if v is not None:
                        parsed_conn_kv[k] = v
                        matched_keys_set.add(k)
        logger().debug(f"kql_engine.py - _inherit_keys - add inherited to combination_keys_set and now inherit_keys_set: {inherit_keys_set}")


    def _check_for_required_keys(self, matched_keys_set:set)->None:
        # make sure that all required keys are in set
        secret_key_set = self._SECRET_KEYS.intersection(self._keys_set)
        missing_set = self._keys_set.difference(matched_keys_set).difference(secret_key_set).difference(self._OPTIONAL_KEYS)
        if len(missing_set) > 0:
            logger().debug(f"kql_engine.py - _check_for_required_keys - the following required key is missing : {missing_set}")
            raise KqlEngineError(f"missing {missing_set}")


    def _check_for_special_combination(self, matched_keys_set:set)->None:
        # special case although tenant in _OPTIONAL_KEYS
        if ConnStrKeys.TENANT not in matched_keys_set and ConnStrKeys.CLIENTID in self._keys_set:
            logger().debug("kql_engine.py - _check_for_special_combination - special case- if clientId exists but tenant doesnt")
            raise KqlEngineError("missing tenant key/value")


    def _check_for_restricted_values(self, matched_keys_set:set, parsed_conn_kv:Dict[str,str])->None:
        # make sure that all required keys are with proper value
        for key in matched_keys_set:  # .difference(secret_key_set).difference(self._SHOULD_BE_NULL_KEYS):
            if key in self._SHOULD_BE_NULL_KEYS:
                if parsed_conn_kv[key] != "":
                    raise KqlEngineError(f"key {key} must be empty")
            elif key not in self._SECRET_KEYS:
                if parsed_conn_kv[key] == f"<{key}>" or parsed_conn_kv[key] == "":
                    raise KqlEngineError(f"key {key} cannot be empty or set to <{key}>")
        logger().debug(f"kql_engine.py - _check_for_restricted_values - make sure that all required keys are with proper value: {matched_keys_set}")


    def _set_and_check_for_cluster_name(self, parsed_conn_kv:Dict[str,str], uri_schema_name:str)->None:
        cluster_name = parsed_conn_kv.get(ConnStrKeys.CLUSTER) or uri_schema_name
        if cluster_name is not None:
            if len(cluster_name) < 1:
                raise KqlEngineError(f"key {ConnStrKeys.CLUSTER} cannot be empty")
            self.cluster_friendly_name = self.createClusterFriendlyName(cluster_name)
        self.cluster_name = cluster_name
        logger().debug(f"kql_engine.py -_set_and_check_for_cluster_name - setting attributes - cluster_name: {self.cluster_name}")
        logger().debug(f"kql_engine.py -_set_and_check_for_cluster_name - setting attributes - cluster_friendly_name: {self.cluster_friendly_name}")


    def _set_and_check_for_database_name(self, parsed_conn_kv:Dict[str,str], mandatory_key:str)->None:
        database_name = parsed_conn_kv.get(mandatory_key)
        if database_name is not None:
            if len(database_name) < 1:
                raise KqlEngineError(f"key {mandatory_key} cannot be empty")
            self.database_friendly_name = self.createDatabaseFriendlyName(database_name)
        self.database_name = database_name
        logger().debug(f"kql_engine.py - _set_and_check_for_database_name - setting attributes - database_name: {self.database_name}")
        logger().debug(f"kql_engine.py - _set_and_check_for_database_name - setting attributes - database_friendly_name: {self.database_friendly_name}")


    def _set_and_check_for_alias(self, parsed_conn_kv:Dict[str,str], friendly_name:str)->None:
        alias = parsed_conn_kv.get(ConnStrKeys.ALIAS) or friendly_name
        if alias is not None:
            if len(alias) < 1 or alias != get_valid_name(alias):
                raise KqlEngineError(f"key {ConnStrKeys.ALIAS} cannot be empty or anything that is not an alphanumeric, dash, underscore, or dot. alias: <{alias}>")
        self.alias = alias
        logger().debug(f"kql_engine.py - _set_and_check_for_alias - setting attributes - alias: {self.alias}")


    def _retreive_and_set_secret(self, matched_keys_set:set, parsed_conn_kv:Dict[str,str])->None:
        # in case secret is missing, get it from user
        secret_key_set = self._SECRET_KEYS.intersection(self._keys_set)
        if len(secret_key_set) == 1:
            s = secret_key_set.pop()
            if s not in matched_keys_set or parsed_conn_kv[s] == f"<{s}>":
                name = self.get_conn_name()
                parsed_conn_kv[s] = getpass.getpass(prompt=f"connection to {name} requires {s}, please enter {s}: ")
                matched_keys_set.add(s)
                logger().debug(f"kql_engine.py - _retreive_and_set_secret - getting secret key from user: {s}")


    def _create_and_set_bind_url(self, parsed_conn_kv:Dict[str,str], uri_schema_name:str)->None:
        items = []
        for key in self._keys_set:
            if key not in self._EXCLUDE_FROM_URL_KEYS:
                items.append(f"{key}('{parsed_conn_kv.get(key)}')")
        self.bind_url = f"{uri_schema_name}://{'.'.join(items)}"
        logger().debug(f"kql_engine.py - _create_and_set_bind_url - setting attributes - bind_url:{self.bind_url}")


    def _parse_common_connection_str(self, conn_str:str, current, uri_schema_name:str, mandatory_key:str, keys_combinations:List[List[str]], user_ns:Dict[str,Any])->Dict[str,str]:
        # logger().debug(f"kql_engine.py -_parse_common_connection_str - params:  conn_str: {conn_str}; current: {current}, uri_schema_name: {uri_schema_name};mandatory_key: {mandatory_key}, valid_keys_combinations: {keys_combinations}, user_ns: {safe_str(user_ns)}")
        logger().debug(f"kql_engine.py -_parse_common_connection_str - params:  conn_str: {conn_str}; current: {current}, uri_schema_name: {uri_schema_name};mandatory_key: {mandatory_key}, valid_keys_combinations: {keys_combinations}")

        try:
            parsed_conn_kv = self._parse_connection_str(conn_str, user_ns)

            matched_keys_set = set(parsed_conn_kv.keys())
            logger().debug(f"kql_engine.py -_parse_common_connection_str - matched_keys_set: {matched_keys_set}")

            self._check_for_unknown_keys(matched_keys_set, keys_combinations)
            self._check_for_mandatory_key(matched_keys_set, mandatory_key)
            self._keys_set = self._find_combination_set(current, matched_keys_set, parsed_conn_kv, keys_combinations)

            self._inherit_keys(current, matched_keys_set, parsed_conn_kv)

            self._check_for_required_keys(matched_keys_set)
            self._check_for_special_combination(matched_keys_set)
            self._check_for_restricted_values(matched_keys_set, parsed_conn_kv)

            self._set_and_check_for_cluster_name(parsed_conn_kv, uri_schema_name)
            self._set_and_check_for_database_name(parsed_conn_kv, mandatory_key)
            self._set_and_check_for_alias(parsed_conn_kv, self.database_friendly_name)

            self._retreive_and_set_secret(matched_keys_set, parsed_conn_kv)

            self._create_and_set_bind_url(parsed_conn_kv, uri_schema_name)

            return parsed_conn_kv
        except Exception as exception:
            raise KqlEngineError(f"invalid connection string: {conn_str}, {exception}")



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

from typing import Union, Dict
import uuid
import json


from .my_utils import json_dumps 
from .constants import Constants, ConnStrKeys, Cloud, Schema
from .kql_response import KqlQueryResponse, KqlSchemaResponse, KqlError
# from .my_aad_helper import _MyAadHelper, ConnKeysKCSB
from .my_aad_helper_msal import _MyAadHelper, ConnKeysKCSB
from ._version import __version__
from .log import logger
from .kql_client import KqlClient
from .exceptions import KqlEngineError

class DraftClient(KqlClient):
    """Draft Client

        Parameters
        ----------
        conn_kv : dict
            Connection string key/value that contains the credentials to access the resource via Draft.
        domain: str
            The Draft client domain, either apps for the case of Application Insights or workspaces for the case Log Analytics.
        data_source: str
            The data source url.
    """

    #
    # Constants
    #


    _DRAFT_CLIENT_BY_CLOUD = {
        Cloud.PUBLIC:      "db662dc1-0cfe-4e1c-a843-19a68e65be58",
        Cloud.MOONCAKE:    "db662dc1-0cfe-4e1c-a843-19a68e65be58",
        Cloud.FAIRFAX:     "730ea9e6-1e1d-480c-9df6-0bb9a90e1a0f",
        Cloud.BLACKFOREST: "db662dc1-0cfe-4e1c-a843-19a68e65be58",
        Cloud.PPE:         "db662dc1-0cfe-4e1c-a843-19a68e65be58",
    }
    _DRAFT_CLIENT_BY_CLOUD[Cloud.CHINA]      = _DRAFT_CLIENT_BY_CLOUD[Cloud.MOONCAKE]
    _DRAFT_CLIENT_BY_CLOUD[Cloud.GOVERNMENT] = _DRAFT_CLIENT_BY_CLOUD[Cloud.FAIRFAX]
    _DRAFT_CLIENT_BY_CLOUD[Cloud.GERMANY]    = _DRAFT_CLIENT_BY_CLOUD[Cloud.BLACKFOREST]

    _WEB_CLIENT_VERSION = __version__
    _API_VERSION = "v1"
    _GET_SCHEMA_QUERY = ".show schema"

    _APPINSIGHTS_URL_BY_CLOUD = {
        Cloud.PUBLIC:      "https://api.applicationinsights.io",
        Cloud.MOONCAKE:    "https://api.applicationinsights.azure.cn",
        Cloud.FAIRFAX:     "https://api.applicationinsights.us",
        Cloud.BLACKFOREST: "https://api.applicationinsights.de",
    }
    _APPINSIGHTS_URL_BY_CLOUD[Cloud.CHINA]      = _APPINSIGHTS_URL_BY_CLOUD[Cloud.MOONCAKE]
    _APPINSIGHTS_URL_BY_CLOUD[Cloud.GOVERNMENT] = _APPINSIGHTS_URL_BY_CLOUD[Cloud.FAIRFAX]
    _APPINSIGHTS_URL_BY_CLOUD[Cloud.GERMANY]    = _APPINSIGHTS_URL_BY_CLOUD[Cloud.BLACKFOREST]

    _LOGANALYTICS_URL_BY_CLOUD = {
        Cloud.PUBLIC:      "https://api.loganalytics.io",
        Cloud.MOONCAKE:    "https://api.loganalytics.azure.cn",
        Cloud.FAIRFAX:     "https://api.loganalytics.us",
        Cloud.BLACKFOREST: "https://api.loganalytics.de",
    }
    _LOGANALYTICS_URL_BY_CLOUD[Cloud.CHINA]      = _LOGANALYTICS_URL_BY_CLOUD[Cloud.MOONCAKE]
    _LOGANALYTICS_URL_BY_CLOUD[Cloud.GOVERNMENT] = _LOGANALYTICS_URL_BY_CLOUD[Cloud.FAIRFAX]
    _LOGANALYTICS_URL_BY_CLOUD[Cloud.GERMANY]    = _LOGANALYTICS_URL_BY_CLOUD[Cloud.BLACKFOREST]

    _DRAFT_URLS_BY_SCHEMA = {
        Schema.APPLICATION_INSIGHTS: _APPINSIGHTS_URL_BY_CLOUD,
        Schema.LOG_ANALYTICS:        _LOGANALYTICS_URL_BY_CLOUD
    }


    def __init__(self, conn_kv:Dict[str,str], domain:str, data_source:str, schema:str, **options)->None:
        super(DraftClient, self).__init__()
        self._domain = domain
        self.resources_name = "workspaces" if schema == Schema.LOG_ANALYTICS else "applications"

        if conn_kv.get(ConnStrKeys.DATA_SOURCE_URL):
            self._data_source =  conn_kv.get(ConnStrKeys.DATA_SOURCE_URL)  
            logger().debug(f"draft_client.py :: __init__ :  self._data_source from conn_kv[\"datasourceurl\"]: {self._data_source}")

        else:
            cloud = options.get("cloud")
            urls = self._DRAFT_URLS_BY_SCHEMA.get(schema, {})
            self._data_source = urls.get(cloud, data_source)
                
            if not self._data_source:
                raise KqlEngineError(f"the service {schema} is not supported in cloud {cloud}")

        logger().debug(f"draft_client.py :: __init__ :  conn_kv[\"datasourceurl\"]: {conn_kv.get(ConnStrKeys.DATA_SOURCE_URL)}")

        self._appkey = conn_kv.get(ConnStrKeys.APPKEY)
        logger().debug(f"draft_client.py :: __init__ :  self._appkey: {self._appkey}")


        if self._appkey is None and conn_kv.get(ConnStrKeys.ANONYMOUS) is None:
            cloud = options.get("cloud")
            client_id = self._DRAFT_CLIENT_BY_CLOUD[cloud]
            http_client = self._http_client if options.get("auth_use_http_client") else None
            self._aad_helper = _MyAadHelper(ConnKeysKCSB(conn_kv, self._data_source), client_id, http_client=http_client, **options)
        else:
            self._aad_helper = None
        logger().debug(f"draft_client.py :: __init__ :  self._aad_helper: {self._aad_helper} ;")


    @property
    def data_source(self)->str:
        return self._data_source


    def execute(self, id:str, query:str, accept_partial_results:bool=False, **options)->Union[KqlQueryResponse, KqlSchemaResponse]:
        """ Execute a simple query or a metadata query
        
        Parameters
        ----------
        id : str
            the workspaces (log analytics) or appid (application insights).
        query : str
            Query to be executed
        accept_partial_results : bool, optional
            Optional parameter. If query fails, but we receive some results, we consider results as partial.
            If this is True, results are returned to client, even if there are exceptions.
            If this is False, exception is raised. Default is False.
        oprions["timeout"] : float, optional
            Optional parameter. Network timeout in seconds. Default is no timeout.

        Returns
        -------
        object
            KqlQueryResponse instnace if executed simple query request
            KqlSchemaResponse instnace if executed metadata request

        Raises
        ------
        KqlError
            If request to draft failed.
            If response from draft contains exceptions.
        """

        #
        # create API url
        #

        is_metadata = query == self._GET_SCHEMA_QUERY
        api_url = f"{self._data_source}/{self._API_VERSION}/{self._domain}/{id}/{'metadata' if is_metadata else 'query'}"

        #
        # create Prefer header
        #

        prefer_list = []
        if self._API_VERSION != "beta":
            prefer_list.append("ai.response-thinning=false")  # returns data as kusto v1
        
        timeout = options.get("timeout")
        if timeout is not None:
            prefer_list.append(f"wait={timeout}")

        #
        # create headers
        #
        
        client_version = f"{Constants.MAGIC_CLASS_NAME}.Python.Client:{self._WEB_CLIENT_VERSION}"        

        client_request_id = f"{Constants.MAGIC_CLASS_NAME}.execute"
        client_request_id_tag = options.get("request_id_tag")
        if client_request_id_tag is not None:
            client_request_id = f"{client_request_id};{client_request_id_tag};{str(uuid.uuid4())}/{self._session_guid}/AzureMonitor"
        else:
            client_request_id = f"{client_request_id};{str(uuid.uuid4())}/{self._session_guid}/AzureMonitor"

        app = f'{Constants.MAGIC_CLASS_NAME};{options.get("notebook_app")}'
        app_tag = options.get("request_app_tag")
        if app_tag is not None:
            app = f"{app};{app_tag}"
            
        user_agent = f'{Constants.MAGIC_CLASS_NAME}/{self._WEB_CLIENT_VERSION}'
        user_agent_tag = options.get("request_user_agent_tag")
        if user_agent_tag is not None:
            user_agent = user_agent_tag

        request_headers = {
            "User-Agent": user_agent,
            "x-ms-client-version": client_version,
            "x-ms-client-request-id": client_request_id,
            "x-ms-app": app
        }
        user_tag = options.get("request_user_tag")
        if user_tag is not None:
            request_headers["x-ms-user"] = user_tag

        if self._aad_helper is not None:
            request_headers["Authorization"] = self._aad_helper.acquire_token()
        elif self._appkey is not None:
            request_headers["x-api-key"] = self._appkey

        if len(prefer_list) > 0:
            request_headers["Prefer"] = ", ".join(prefer_list)

        cache_max_age = options.get("request_cache_max_age")
        if cache_max_age is not None:
            if cache_max_age > 0:
                request_headers["Cache-Control"] = f"max-age={cache_max_age}"
            else:
                request_headers["Cache-Control"] = "no-cache"

        #
        # submit request
        #
        log_request_headers = request_headers
        if request_headers.get("Authorization"):
            log_request_headers = request_headers.copy()
            log_request_headers["Authorization"] = "..." 

        # collect this inormation, in case bug report will be generated
        KqlClient.last_query_info = {
            "request": {
                "endpoint": api_url,
                "headers": log_request_headers,
                "timeout": options.get("timeout"),
            }
        }

        if is_metadata:
            logger().debug(f"DraftClient::execute - GET request - url: {api_url}, headers: {log_request_headers}, timeout: {options.get('timeout')}")
            response = self._http_client.get(api_url, headers=request_headers, timeout=options.get("timeout"))
        else:
            request_payload = {
                "query": query
            }

            # Implicit Cross Workspace Queries: https://dev.loganalytics.io/oms/documentation/3-Using-the-API/CrossResourceQuery
            # workspaces - string[] - A list of workspaces that are included in the query.
            if type(options.get("query_properties")) == dict:
                resources = options.get("query_properties").get(self.resources_name)
                if type(resources) == list and len(resources) > 0:
                    request_payload[self.resources_name] = resources

                timespan = options.get("query_properties").get("timespan")
                if type(timespan) == str and len(timespan) > 0:
                    request_payload["timespan"] = timespan

            logger().debug(f"DraftClient::execute - POST request - url: {api_url}, headers: {log_request_headers}, payload: {request_payload}, timeout: {options.get('timeout')}")

            # collect this inormation, in case bug report will be generated
            self.last_query_info["request"]["payload"] = request_payload  # pylint: disable=unsupported-assignment-operation, unsubscriptable-object

            response = self._http_client.post(api_url, headers=request_headers, json=request_payload, timeout=options.get("timeout"))

        logger().debug(f"DraftClient::execute - response - status: {response.status_code}, headers: {response.headers}, payload: {response.text}")
        #
        # handle response
        #
        # collect this inormation, in case bug report will be generated
        self.last_query_info["response"] = {  # pylint: disable=unsupported-assignment-operation
            "status_code": response.status_code
        }

        if response.status_code < 200  or response.status_code >= 300:  # pylint: disable=E1101
            try:
                parsed_error = json.loads(response.text)
            except:
                parsed_error = response.text
            # collect this inormation, in case bug report will be generated
            self.last_query_info["response"]["error"] = parsed_error  # pylint: disable=unsupported-assignment-operation, unsubscriptable-object
            raise KqlError(response.text, response)

        json_response = response.json()

        if is_metadata:
            kql_response = KqlSchemaResponse(json_response)
        else:
            kql_response = KqlQueryResponse(json_response)

        if kql_response.has_exceptions() and not accept_partial_results:
            try:
                error_message = json_dumps(kql_response.get_exceptions())
            except:
                error_message = str(kql_response.get_exceptions())
            raise KqlError(error_message, response, kql_response)

        return kql_response



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import os
from typing import Any, Union, Dict


from .kql_engine import KqlEngine 
from .cache_client import CacheClient
from .constants import ConnStrKeys
from .exceptions import KqlEngineError
from .my_utils import get_valid_filename_with_spaces, adjust_path


class CacheEngine(KqlEngine):

    _URI_SCHEMA_NAME = "cache"  # no spaces, underscores, and hyphe-minus, because they are ignored in parser
    _ALT_URI_SCHEMA_NAME = "file"  # no spaces, underscores, and hyphe-minus, because they are ignored in parser

    _ALT_URI_SCHEMA_NAMES = [_URI_SCHEMA_NAME, _ALT_URI_SCHEMA_NAME]
    _RESERVED_CLUSTER_NAME = _URI_SCHEMA_NAME
    _MANDATORY_KEY = ConnStrKeys.FOLDER
    _VALID_KEYS_COMBINATIONS = [[ConnStrKeys.FOLDER, ConnStrKeys.ALIAS]]

    _VALIDATION_FILE_NAME = get_valid_filename_with_spaces("validation_file.json")


    # Object constructor
    def __init__(self, conn_str_or_engine:Union[str,KqlEngine], user_ns:Dict[str,Any], current:KqlEngine=None, cache_name:str=None, **options)->None:
        super(CacheEngine, self).__init__()
        self._parsed_conn = {}
        self.kql_engine = None
        if isinstance(conn_str_or_engine, KqlEngine):
            self.kql_engine = conn_str_or_engine
            folder_name = f"{self.kql_engine.get_database_friendly_name()}_at_{self.kql_engine.cluster_friendly_name()}"
            conn_str = f"{self._URI_SCHEMA_NAME}://{ConnStrKeys.FOLDER}='{folder_name}'"
        else:
            conn_str = conn_str_or_engine

        self._parsed_conn = self._parse_common_connection_str(
            conn_str, current, self._URI_SCHEMA_NAME, self._MANDATORY_KEY, self._VALID_KEYS_COMBINATIONS, user_ns
        )
        self.client = CacheClient(**options)

        folder_path = self.client._get_folder_path(self.get_database_friendly_name(), cache_name)
        validation_file_path = adjust_path(f"{folder_path}/{self._VALIDATION_FILE_NAME}")
        if not os.path.exists(validation_file_path):
            outfile = open(validation_file_path, "w")
            outfile.write(self.validate_json_file_content)
            outfile.flush()
            outfile.close()


    def validate(self, **options)->None:
        # query = "range c from 1 to 10 step 1 | count"
        filename = self._VALIDATION_FILE_NAME
        database = self.get_database_friendly_name()
        kql_response = self.execute(database, filename, accept_partial_results=False, **options)
        table = kql_response.tables[0]
        if table.rowcount() != 1 or table.colcount() != 1 or [r for r in table.fetchall()][0][0] != 10:
            raise KqlEngineError("Client failed to validate connection.")


    validate_json_file_content = """{"Tables": [{"TableName": "Table_0", "Columns": [{"ColumnName": "Count", "DataType": "Int64", "ColumnType": "long"}], "Rows": [[10]]}, {"TableName": "Table_1", "Columns": [{"ColumnName": "Value", "DataType": "String", "ColumnType": "string"}], "Rows": [["{\\"Visualization\\":null,\\"Title\\":null,\\"XColumn\\":null,\\"Series\\":null,\\"YColumns\\":null,\\"XTitle\\":null,\\"YTitle\\":null,\\"XAxis\\":null,\\"YAxis\\":null,\\"Legend\\":null,\\"YSplit\\":null,\\"Accumulate\\":false,\\"IsQuerySorted\\":false,\\"Kind\\":null}"]]}, {"TableName": "Table_2", "Columns": [{"ColumnName": "Timestamp", "DataType": "DateTime", "ColumnType": "datetime"}, {"ColumnName": "Severity", "DataType": "Int32", "ColumnType": "int"}, {"ColumnName": "SeverityName", "DataType": "String", "ColumnType": "string"}, {"ColumnName": "StatusCode", "DataType": "Int32", "ColumnType": "int"}, {"ColumnName": "StatusDescription", "DataType": "String", "ColumnType": "string"}, {"ColumnName": "Count", "DataType": "Int32", "ColumnType": "int"}, {"ColumnName": "RequestId", "DataType": "Guid", "ColumnType": "guid"}, {"ColumnName": "ActivityId", "DataType": "Guid", "ColumnType": "guid"}, {"ColumnName": "SubActivityId", "DataType": "Guid", "ColumnType": "guid"}, {"ColumnName": "ClientActivityId", "DataType": "String", "ColumnType": "string"}], "Rows": [["2018-09-17T01:45:07.5325114Z", 4, "Info", 0, "Query completed successfully", 1, "21d61568-0a1a-41e2-ab8c-7a85992a1f3b", "21d61568-0a1a-41e2-ab8c-7a85992a1f3b", "8a9c6cc6-f723-431f-9396-4c91ec9a8837", "9dff54f7-dd4c-445f-89e1-02b50661086e"], ["2018-09-17T01:45:07.5325114Z", 6, "Stats", 0, "{\\"ExecutionTime\\":0.0,\\"resource_usage\\":{\\"cache\\":{\\"memory\\":{\\"hits\\":0,\\"misses\\":0,\\"total\\":0},\\"disk\\":{\\"hits\\":0,\\"misses\\":0,\\"total\\":0}},\\"cpu\\":{\\"user\\":\\"00:00:00\\",\\"kernel\\":\\"00:00:00\\",\\"total cpu\\":\\"00:00:00\\"},\\"memory\\":{\\"peak_per_node\\":0}},\\"input_dataset_statistics\\":{\\"extents\\":{\\"total\\":0,\\"scanned\\":0},\\"rows\\":{\\"total\\":0,\\"scanned\\":0}},\\"dataset_statistics\\":[{\\"table_row_count\\":1,\\"table_size\\":8}]}", 1, "21d61568-0a1a-41e2-ab8c-7a85992a1f3b", "21d61568-0a1a-41e2-ab8c-7a85992a1f3b", "8a9c6cc6-f723-431f-9396-4c91ec9a8837", "9dff54f7-dd4c-445f-89e1-02b50661086e"]]}, {"TableName": "Table_3", "Columns": [{"ColumnName": "Ordinal", "DataType": "Int64", "ColumnType": "long"}, {"ColumnName": "Kind", "DataType": "String", "ColumnType": "string"}, {"ColumnName": "Name", "DataType": "String", "ColumnType": "string"}, {"ColumnName": "Id", "DataType": "String", "ColumnType": "string"}, {"ColumnName": "PrettyName", "DataType": "String", "ColumnType": "string"}], "Rows": [[0, "QueryResult", "PrimaryResult", "1bd5362f-e1f6-4258-abb3-9c2fedca8bdb", ""], [1, "QueryProperties", "@ExtendedProperties", "b1f9ef32-f6f7-4304-9e94-616a3472fb7e", ""], [2, "QueryStatus", "QueryStatus", "00000000-0000-0000-0000-000000000000", ""]]}]}"""



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

from datetime import timedelta
import re
import json
from decimal import Decimal
import collections


import dateutil.parser


from .constants import ExtendedPropertiesKeys


try:
    collectionsAbc = collections.abc
except AttributeError:
    collectionsAbc = collections



# Regex for TimeSpan
_TIMESPAN_PATTERN = re.compile(r"(-?)((?P<d>[0-9]*).)?(?P<h>[0-9]{2}):(?P<m>[0-9]{2}):(?P<s>[0-9]{2}(\.[0-9]+)?$)")


class KqlResult(dict):
    """ Simple wrapper around dictionary, to enable both index and key access to rows in result """

    def __init__(self, index2column_mapping, *args, **kwargs):
        super(KqlResult, self).__init__(*args)
        # TODO: this is not optimal, if client will not access all fields.
        # In that case, we are having unnecessary perf hit to convert Timestamp, even if client don't use it.
        # In this case, it would be better for KqlResult to extend list class. In this case,
        # KqlResultIter.index2column_mapping should be reversed, e.g. column2index_mapping.
        self.index2column_mapping = index2column_mapping


    def __getitem__(self, key):
        if isinstance(key, slice):
            start = min(key.start or 0, len(self))
            end = min(key.stop or len(self), len(self))
            mapping = self.index2column_mapping[start:end]
            dic = dict([(c, dict.__getitem__(self, c)) for c in mapping])
            return KqlResult(mapping, dic)
        elif isinstance(key, int):
            val = dict.__getitem__(self, self.index2column_mapping[key])
        else:
            val = dict.__getitem__(self, key)
        return val


class KqlResponseTable(collectionsAbc.Iterator):
    """ Iterator over returned rows """

    def __init__(self, id, response_table):
        self.id = id
        self.rows = response_table["Rows"]
        self.columns = response_table["Columns"]
        self.index2column_mapping = []
        self.index2type_mapping = []
        for c in self.columns:
            self.index2column_mapping.append(c["ColumnName"])
            ctype = c["ColumnType"] if "ColumnType" in c else c["DataType"]
            self.index2type_mapping.append(ctype)
        self.row_index = 0
        self._rows_count = sum([1 for r in self.rows if isinstance(r,list)])  # len(self.rows)
        # Here we keep converter functions for each type that we need to take special care (e.g. convert)

        # index MUST be lowercase !!!
        self.converters_lambda_mappings = {
            "datetime": self.to_datetime,
            "timespan": self.to_timedelta,
            "decimal": self.to_decimal,
            "dynamic": self.to_object,
        }


    @staticmethod
    def to_object(value):
        try:
            return json.loads(value) if value and isinstance(value, str) else value
        except Exception:
            return value


    @staticmethod
    def to_datetime(value):
        """Converts a string/int to a datetime."""
        if value is None:
            return None

        if isinstance(value, int):
            return dateutil.parser.parse(str(value))

        return dateutil.parser.isoparse(value)


    @staticmethod
    def to_decimal(value):
        """Converts a decimal string to Decimal."""

        if value is None:
            return None

        return Decimal(value)


    @staticmethod
    def to_timedelta(value):
        """Converts a string/int/float to a timedelta."""
        if value is None:
            return None

        if isinstance(value, (int, float)):
            return timedelta(microseconds=(float(value) / 10))

        match = _TIMESPAN_PATTERN.match(value)

        if match:
            if match.group(1) == "-":
                factor = -1
            else:
                factor = 1
            return factor * timedelta(
                days=int(match.group("d") or 0), hours=int(match.group("h")), minutes=int(match.group("m")), seconds=float(match.group("s"))
            )
        else:
            raise ValueError(f"Timespan value '{value}' cannot be decoded")


    def __iter__(self):
        self.row_index = 0
        return self


    def __next__(self):
        if self.row_index >= self.rows_count:
            raise StopIteration
        row = self.rows[self.row_index]
        result_dict = {}
        for index, value in enumerate(row):
            data_type = self.index2type_mapping[index].lower()
            column_name = self.index2column_mapping[index]
            if data_type in self.converters_lambda_mappings:
                value = self.converters_lambda_mappings[data_type](value)
            elif self.rows_count == 1 and self.columns_count == 1 and column_name == "DatabaseSchema" and data_type == "string":
                value = self.to_object(value)
            result_dict[column_name] = value
        self.row_index = self.row_index + 1
        return KqlResult(self.index2column_mapping, result_dict)


    @property
    def columns_name(self):
        return self.index2column_mapping


    @property
    def is_partial(self):
        return len(self.rows) > self._rows_count


    @property
    def columns_type(self):
        return self.index2type_mapping


    @property
    def rows_count(self):
        return self._rows_count


    @property
    def columns_count(self):
        return len(self.columns)


    def fetchall(self):
        """ Returns iterator to get rows from response """
        return self.__iter__()


    def iter_all(self):
        """ Returns iterator to get rows from response """
        return self.__iter__()


class KqlSchemaResponse(object):

    def __init__(self, json_response):
        self.json_response = json_response
        self.table = json_response["tables"]


    def has_exceptions(self):
        return "Exceptions" in self.json_response


    def get_exceptions(self):
        return self.json_response["Exceptions"]


class KqlQueryResponse(object):
    """ Wrapper for response """

    # TODO: add support to get additional infromation from response, like execution time

    def __init__(self, json_response, endpoint_version="v1"):
        self.json_response = json_response
        self.endpoint_version = "v1" if not isinstance(self.json_response, list) else endpoint_version
        self.visualization = None
        self._extended_properties = None

        if self.endpoint_version == "v2":
            self.all_tables = [t for t in json_response if t["FrameType"] == "DataTable"]
            self.tables = [t for t in json_response if t["FrameType"] == "DataTable" and t["TableKind"] == "PrimaryResult"]
            self.primary_results = [KqlResponseTable(t["TableId"], t) for t in self.tables]
            self.dataSetCompletion = [f for f in json_response if f["FrameType"] == "DataSetCompletion"]
        else:
            self.all_tables = self.json_response["Tables"]
            tables_num = self.json_response["Tables"].__len__()
            last_table = self.json_response["Tables"][tables_num - 1]
            if tables_num < 2:
                self.tables = []
            else:
                self.tables = [self.json_response["Tables"][r[0]] for r in last_table["Rows"] if r[2] == "GenericResult" or r[2] == "PrimaryResult"]
            if len(self.tables) == 0:
                self.tables = self.all_tables[:1]
            self.primary_results = [KqlResponseTable(idx, t) for idx, t in enumerate(self.tables)]
            self.dataSetCompletion = []
 

    def _get_endpoint_version(self, json_response):
        try:
            tables_num = json_response["Tables"].__len__()  # pylint: disable=W0612
            return "v1"
        except:
            return "v2"


    @property
    def extended_properties(self):
        if self._extended_properties is None:
            self._extended_properties = {}
            if self.endpoint_version == "v2":
                for table in self.all_tables:
                    if table["TableName"] == "@ExtendedProperties" and table["TableKind"] == "QueryProperties":
                        cols_idx_map = self._map_columns_to_index(table["Columns"])
                        types = self._get_columns_types(table["Columns"])
                        key_idx = cols_idx_map.get("Key")
                        table_id_idx = cols_idx_map.get("TableId")
                        value_idx = cols_idx_map.get("Value")
                        if (
                            key_idx is not None
                            and table_id_idx is not None
                            and value_idx is not None
                            and types[key_idx] == "string"
                            and types[table_id_idx] == "int"
                        ):
                            for row in table["Rows"]:
                                table_id = row[table_id_idx]
                                key = row[key_idx]
                                value = row[value_idx]
                                extended_properties = self._extended_properties.get(table_id, {})
                                if key == ExtendedPropertiesKeys.VISUALIZATION:
                                    value_obj = self._dynamic_to_object(value)
                                    extended_properties[key] = value_obj
                                elif key == ExtendedPropertiesKeys.CURSOR:
                                    # Note: although the value column is defined as dynamic the value here is string
                                    extended_properties[key] = str(value)
                                else:
                                    extended_properties[key] = value
                                self._extended_properties[table_id] = extended_properties

            else:
                tables_num = self.json_response["Tables"].__len__()
                if tables_num > 1:
                    last_table = self.json_response["Tables"][tables_num - 1]
                    for row in last_table["Rows"]:
                        if row[2] == "@ExtendedProperties" and row[1] == "QueryProperties":
                            table = self.json_response["Tables"][row[0]]
                            table_id = 0
                            value_idx = 0
                            extended_properties = {}
                            for row in table["Rows"]:
                                value = row[value_idx]
                                value_obj = self._dynamic_to_object(value)
                                if type(value_obj) == dict:
                                    key = list(value_obj.keys())[0]
                                    if key == ExtendedPropertiesKeys.VISUALIZATION:
                                        extended_properties[key] = value_obj
                                    elif key == ExtendedPropertiesKeys.CURSOR:
                                        extended_properties[key] = str(value_obj.get(ExtendedPropertiesKeys.CURSOR,""))
                                    else:
                                        extended_properties[key] = value_obj
                                    self._extended_properties[table_id] = extended_properties
                            break
        return self._extended_properties


    @property
    def completion_query_info_results(self):
        if self.endpoint_version == "v2":
            for table in self.all_tables:
                if table["TableName"] == "QueryCompletionInformation":
                    cols_idx_map = self._map_columns_to_index(table["Columns"])
                    event_type_name_idx = cols_idx_map.get("EventTypeName")
                    payload_idx = cols_idx_map.get("Payload")
                    if event_type_name_idx is not None and payload_idx is not None:
                        for row in table["Rows"]:
                            if row[event_type_name_idx] == "QueryInfo":
                                value = row[payload_idx]
                                return self._dynamic_to_object(value)
        else:
            tables_num = self.json_response["Tables"].__len__()
            if tables_num > 1:
                last_table = self.json_response["Tables"][tables_num - 1]
                for r in last_table["Rows"]:
                    if r[2] == "QueryStatus":
                        t = self.json_response["Tables"][r[0]]
                        for sr in t["Rows"]:
                            if sr[2] == "Info":
                                info = {"StatusCode": sr[3], "StatusDescription": sr[4], "Count": sr[5]}
                                # print(f'Info: {info}')
                                return info
        return {}


    @property
    def completion_query_resource_consumption_results(self):
        if self.endpoint_version == "v2":
            for table in self.all_tables:
                if table["TableName"] == "QueryCompletionInformation":
                    cols_idx_map = self._map_columns_to_index(table["Columns"])
                    event_type_name_idx = cols_idx_map.get("EventTypeName")
                    payload_idx = cols_idx_map.get("Payload")
                    if event_type_name_idx is not None and payload_idx is not None:
                        for row in table["Rows"]:
                            if row[event_type_name_idx] == "QueryResourceConsumption":
                                value = row[payload_idx]
                                return self._dynamic_to_object(value)
        else:
            tables_num = self.json_response["Tables"].__len__()
            if tables_num > 1:
                last_table = self.json_response["Tables"][tables_num - 1]
                for r in last_table["Rows"]:
                    if r[2] == "QueryStatus":
                        t = self.json_response["Tables"][r[0]]
                        for sr in t["Rows"]:
                            if sr[2] == "Stats":
                                stats = sr[4]
                                # print(f'stats: {stats}')
                                return self._dynamic_to_object(stats)
        return {}


    @property
    def dataSetCompletion_results(self):
        return self.dataSetCompletion


    def _map_columns_to_index(self, columns: list):
        map = {}
        for idx, col in enumerate(columns):
            map[col["ColumnName"]] = idx
        return map


    def _get_columns_types(self, columns: list):
        map = []
        for col in columns:
            map.append(col["ColumnType"])
        return map


    def get_raw_response(self):
        return self.json_response


    def get_table_count(self):
        return len(self.tables)


    def has_exceptions(self):
        return "Exceptions" in self.json_response


    def get_exceptions(self):
        return self.json_response["Exceptions"]


    @staticmethod
    def _dynamic_to_object(value:str):
        try:
            return json.loads(value) if value and isinstance(value, str) else value if value else None
        except Exception:
            return value


class KqlError(Exception):
    """
    Represents error returned from server. Error can contain partial results of the executed query.
    """

    def __init__(self, message, http_response, kql_response=None):
        super(KqlError, self).__init__(message)
        self.message = message
        self.http_response = http_response
        self.kql_response = kql_response


    def get_raw_http_response(self):
        return self.http_response


    def is_semantic_error(self):
        return self.http_response.text.startswith("Semantic error:")


    def has_partial_results(self):
        return self.kql_response is not None


    def get_partial_results(self):
        return self.kql_response



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

from typing import Any, List, Dict
import re


from .constants import Constants
from .display import Display
from .kql_engine import KqlEngine
from .kusto_engine import KustoEngine
from .aria_engine import AriaEngine
from .ai_engine import AppinsightsEngine
from .aimon_engine import AimonEngine
from .la_engine import LoganalyticsEngine
from .cache_engine import CacheEngine
from .cache_client import CacheClient


class Database_html(object):
    """
    """

    _database_metadata_css = """.just-padding {
      height: 100%;
      width: 100%;
      padding: 15px;
    }

    .list-group.list-group-root {
      padding: 0;
      overflow: hidden;
    }

    .list-group.list-group-root .list-group {
      margin-bottom: 0;
    }

    .list-group.list-group-root .list-group-item {
      border-radius: 0;
      border-width: 1px 0 0 0;
    }

    .list-group.list-group-root > .list-group-item:first-child {
      border-top-width: 0;
    }

    .list-group.list-group-root > .list-group > .list-group-item {
      padding-left: 60px;
    }

    .list-group.list-group-root > .list-group > .list-group > .list-group-item {
      padding-left: 60px;
    }"""

    _database_metadata_scripts = """
        <link href="https://maxcdn.bootstrapcdn.com/bootstrap/3.3.5/css/bootstrap.min.css" rel="stylesheet" type="text/css">
        <script src="https://ajax.googleapis.com/ajax/libs/jquery/1.11.3/jquery.min.js" type="text/javascript"></script>
        <script src="https://maxcdn.bootstrapcdn.com/bootstrap/3.3.5/js/bootstrap.min.js" type="text/javascript"></script>
                <script type="text/javascript">
                    var w = screen.width;
                    var h = screen.height;            
                    window.resizeTo(w/4, h);
                    window.focus(); 
                </script>
        <script type="text/javascript">
            window.onload=function(){
      
        $(function() {

          $('.list-group-item').on('click', function() {
            $('.glyphicon', this)
              .toggleClass('glyphicon-chevron-right')
              .toggleClass('glyphicon-chevron-down');
          });

        });

            }
        </script>
    """

    _database_metadata_html = """<!DOCTYPE html>
        <html><head>
        <meta charset="utf-8">
        <meta name="viewport" content="width=device-width,initial-scale=1">
        <title>{0}</title>

        </head><body>
        {1}
        <style>
        {2}
        </style>

        <h1 align="center">{3}</h1>

        <div class="just-padding">

        <div class="list-group list-group-root well">
        {4}
        </div></div></body></html>"""


    @classmethod
    def _convert_database_metadata_to_html(cls, database_metadata_tree:Dict[str,Any], conn_name:str, **kwargs)->str:
        item = ""
        for table_name in database_metadata_tree.keys():
            table_metadata_tree = database_metadata_tree.get(table_name)
            item += cls._convert_table_metadata_tree_to_item(table_name, table_metadata_tree, **kwargs)
        header = conn_name
        title = f"{Constants.MAGIC_PACKAGE_NAME} - {conn_name.replace('@', '_at_')} schema"
        result = cls._database_metadata_html.format(
            title, cls._database_metadata_scripts, cls._database_metadata_css, header, item
        )
        return result


    @classmethod
    def _create_database_metadata_tree(cls, rows, database_name:str, **kwargs)->Dict[str,Any]:
        database_metadata_tree = {}
        for row in rows:
            database_name:str = row["DatabaseName"]
            table_name = row["TableName"]
            column_name = row["ColumnName"]
            column_type = row["ColumnType"]
            if database_name.lower() == database_name.lower():
                if table_name and len(table_name) > 0:
                    if not database_metadata_tree.get(table_name):
                        database_metadata_tree[table_name] = {}
                    if column_name and len(column_name) > 0 and column_type and len(column_type) > 0:
                        database_metadata_tree.get(table_name)[column_name] = column_type
        return database_metadata_tree


    @classmethod
    def _create_database_draft_metadata_tree(cls, rows:List[Dict[str,Any]], **kwargs)->Dict[str,Any]:
        database_metadata_tree = {}
        for row in rows:
            table_name = row["name"]
            if table_name and len(table_name) > 0:
                database_metadata_tree[table_name] = {}
                for col in row["columns"]:
                    column_name = col["name"]
                    column_type = col["type"]
                    if column_name and len(column_name) > 0 and column_type and len(column_type) > 0:
                        database_metadata_tree.get(table_name)[column_name] = column_type
        return database_metadata_tree


    @classmethod
    def _convert_table_metadata_tree_to_item(cls, table_name:str, table_metadata_tree:Dict[str,Any], **kwargs):
        metadata_items = []
        for column_name in table_metadata_tree.keys():
            column_type = table_metadata_tree.get(column_name)
            if column_type.startswith("System."):
                column_type = column_type[7:]
            metadata_item = cls._convert_column_metadata_to_item(column_name, column_type, **kwargs)
            metadata_items.append(metadata_item)
            
        item = (
            f"""<a href='#{table_name}' class="list-group-item" data-toggle="collapse">
                     <i class="glyphicon glyphicon-chevron-right"></i><b>{table_name}</b></a>
                  <div class="list-group collapse" id='{table_name}'>{''.join(metadata_items)}</div>"""
        )
        return item


    @classmethod
    def _convert_column_metadata_to_item(cls, column_name:str, column_type:str, **kwargs)->str:
        item = f"<b>{column_name}</b> : {column_type}"
        return f"""<a href="#" class="list-group-item">{item}</a>"""


    @classmethod
    def get_schema_tree(cls, engine:KqlEngine, **options)->Dict[str,Any]:
        engine_class = (
            AriaEngine
            if isinstance(engine, AriaEngine) or (isinstance(engine, CacheEngine) and isinstance(engine.kql_engine, AriaEngine))
            # must be after AriaEngine, because AriaEngine class inherit from KustoEngine
            else KustoEngine
            if isinstance(engine, KustoEngine) or (isinstance(engine, CacheEngine) and isinstance(engine.kql_engine, KustoEngine))
            else LoganalyticsEngine
            if isinstance(engine, LoganalyticsEngine) or (isinstance(engine, CacheEngine) and isinstance(engine.kql_engine, LoganalyticsEngine))
            else AimonEngine
            if isinstance(engine, AimonEngine) or (isinstance(engine, CacheEngine) and isinstance(engine.kql_engine, AimonEngine))
            # must be after AimonEngine, because AimonEngine class inherit from AppinsightsEngine
            else AppinsightsEngine 
            if isinstance(engine, AppinsightsEngine) or (isinstance(engine, CacheEngine) and isinstance(engine.kql_engine, AppinsightsEngine))
            else None
        )

        if engine_class is not None:
            if isinstance(engine, CacheEngine):
                client_database_name = engine.kql_engine.get_client_database_name()
            else:
                client_database_name = engine.get_client_database_name()

            if engine_class in [KustoEngine, AriaEngine]:
                show_schema_query = f".show database ['{cls._adjustToKustoEntityNameRules(client_database_name)}'] schema"
                raw_query_result = engine.execute(show_schema_query, **options)
                raw_schema_table = raw_query_result.tables[0]
                database_metadata_tree = cls._create_database_metadata_tree(raw_schema_table.fetchall(), client_database_name)
                if options.get("cache") is not None and options.get("cache") != options.get("use_cache"):
                    CacheClient(**options).save(raw_query_result, engine, show_schema_query, **options)
                return database_metadata_tree

            elif engine_class in [AppinsightsEngine, LoganalyticsEngine, AimonEngine]:
                show_schema_query = ".show schema"
                metadata_result = engine.client_execute(show_schema_query, **options)
                metadata_schema_table = metadata_result.table
                database_metadata_tree = cls._create_database_draft_metadata_tree(metadata_schema_table)
                if options.get("cache") is not None and options.get("cache") != options.get("use_cache"):
                    CacheClient(**options).save(metadata_result, engine, show_schema_query, **options)
                return database_metadata_tree
        return None


    @classmethod
    def _adjustToKustoEntityNameRules(cls, name:str)->str:
        if isinstance(name, str):
            name = re.sub(r'[\s\n\r\f\t]+', ' ', name.strip())
            name = re.sub(r'[^0-9a-zA-Z._\s-]+', ' ', name)
        return name


    @classmethod
    def get_schema_file_path(cls, engine:KqlEngine, **options)->None:
        database_metadata_tree = cls.get_schema_tree(engine, **options)
        if database_metadata_tree is not None:
            if isinstance(engine, CacheEngine):
                conn_name = engine.kql_engine.get_conn_name()
            else:
                conn_name = engine.get_conn_name()
            html_str = cls._convert_database_metadata_to_html(database_metadata_tree, conn_name)
            window_name = f"_{conn_name.replace('@', '_at_')}_schema"
            return Display._html_to_file_path(html_str, window_name, **options)
        else:
            return None


    @classmethod
    def popup_schema(cls, file_path:str, engine:KqlEngine, **options)->None:
        if file_path:
            conn_name = engine.kql_engine.get_conn_name() if isinstance(engine, CacheEngine) else engine.get_conn_name()
            button_text = f"popup schema {conn_name}"
            window_name = f"_{conn_name.replace('@', '_at_')}_schema"
            Display.show_window(window_name, file_path, button_text=button_text, onclick_visibility="visible", content="schema", **options)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

from typing import Any, Dict


from .ai_engine import AppinsightsEngine
from .constants import Schema


class AimonEngine(AppinsightsEngine):

    # Constants
    # ---------
    _URI_SCHEMA_NAME = Schema.AIMON  # no spaces, underscores, and hyphe-minus, because they are ignored in parser
    _ALT_URI_SCHEMA_NAME = Schema.AIMON  # no spaces, underscores, and hyphe-minus, because they are ignored in parser
    _DOMAIN = "apps"
    
    _DATA_SOURCE = "https://api.aimon.applicationinsights.io"
 
    _ALT_URI_SCHEMA_NAMES = [_URI_SCHEMA_NAME]
    _RESERVED_CLUSTER_NAME = _URI_SCHEMA_NAME
    
    # Class methods
    # -------------

            
    # Instance methods
    # ----------------

    def __init__(self, conn_str:str, user_ns:Dict[str,Any], current:AppinsightsEngine=None, **options)->None:
        super(AimonEngine, self).__init__(conn_str, user_ns, current, **options)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

from typing import Union
import uuid

from .kql_response import KqlQueryResponse, KqlSchemaResponse
from .aad_helper import AadHelper
from .http_client import HttpClient


class KqlClient(object):

    # collect this information, in case bug report will be generated
    last_query_info:dict = None

    _aad_helper:AadHelper
    _http_client:HttpClient

    _session_guid:str = str(uuid.uuid4())


    def __init__(self)->None:
        self._aad_helper = None
        self._http_client = HttpClient()


    def execute(self, id:str, query:str, accept_partial_results:bool=False, **options)->Union[KqlQueryResponse, KqlSchemaResponse]:
        raise NotImplementedError(self.__class__.__name__ + ".execute")



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

from typing import Any, Dict


from .kusto_engine import KustoEngine
from .constants import ConnStrKeys, Schema


class AriaEngine(KustoEngine):

    # Constants
    # ---------

    _URI_SCHEMA_NAME = Schema.ARIA  # no spaces, underscores, and hyphe-minus, because they are ignored in parser

    _ALT_URI_SCHEMA_NAMES = [_URI_SCHEMA_NAME]
    _RESERVED_CLUSTER_NAME = _URI_SCHEMA_NAME

    _DEFAULT_CLUSTER_NAME = "https://kusto.aria.microsoft.com"

    _VALID_KEYS_COMBINATIONS = [
        [ConnStrKeys.DATABASE, ConnStrKeys.ALIAS, ConnStrKeys.TENANT, ConnStrKeys.AAD_URL, ConnStrKeys.CLIENTID, ConnStrKeys.CLIENTSECRET],
        [ConnStrKeys.DATABASE, ConnStrKeys.ALIAS, ConnStrKeys.TENANT, ConnStrKeys.AAD_URL, ConnStrKeys.CLIENTID, ConnStrKeys.CERTIFICATE, ConnStrKeys.CERTIFICATE_THUMBPRINT],
        [ConnStrKeys.DATABASE, ConnStrKeys.ALIAS, ConnStrKeys.TENANT, ConnStrKeys.AAD_URL, ConnStrKeys.CLIENTID, ConnStrKeys.CODE],
        [ConnStrKeys.DATABASE, ConnStrKeys.ALIAS, ConnStrKeys.TENANT, ConnStrKeys.AAD_URL,                       ConnStrKeys.CODE],
        [ConnStrKeys.DATABASE, ConnStrKeys.ALIAS, ConnStrKeys.TENANT, ConnStrKeys.AAD_URL, ConnStrKeys.CLIENTID, ConnStrKeys.USERNAME, ConnStrKeys.PASSWORD],
        [ConnStrKeys.DATABASE, ConnStrKeys.ALIAS, ConnStrKeys.TENANT, ConnStrKeys.AAD_URL,                       ConnStrKeys.USERNAME, ConnStrKeys.PASSWORD],
        [ConnStrKeys.DATABASE, ConnStrKeys.ALIAS,                                                                ConnStrKeys.ANONYMOUS],
    ]

    _VALID_KEYS_COMBINATIONS_NEW = [
        {
            "must": [ConnStrKeys.DATABASE, ConnStrKeys.CLIENTID, ConnStrKeys.CLIENTSECRET],
            "extra": [],
            "optional": [ConnStrKeys.ALIAS, ConnStrKeys.TENANT, ConnStrKeys.AAD_URL]
        },
        {
            "must": [ConnStrKeys.DATABASE, ConnStrKeys.CLIENTID, ConnStrKeys.CERTIFICATE, ConnStrKeys.CERTIFICATE_THUMBPRINT],
            "extra": [],
            "optional": [ConnStrKeys.ALIAS, ConnStrKeys.TENANT, ConnStrKeys.AAD_URL]
        },
        {
            "must": [ConnStrKeys.DATABASE, ConnStrKeys.CODE],
            "extra": [ConnStrKeys.CLIENTID, ConnStrKeys.USERNAME],
            "optional": [ConnStrKeys.ALIAS, ConnStrKeys.TENANT, ConnStrKeys.AAD_URL]
        },
        {
            "must": [ConnStrKeys.DATABASE, ConnStrKeys.USERNAME, ConnStrKeys.PASSWORD],
            "extra": [ConnStrKeys.CLIENTID],
            "optional": [ConnStrKeys.ALIAS, ConnStrKeys.TENANT, ConnStrKeys.AAD_URL, ConnStrKeys.CLIENTID]
        },
        {
            "must": [ConnStrKeys.DATABASE, ConnStrKeys.ANONYMOUS],
            "extra": [],
            "optional": [ConnStrKeys.ALIAS]
        }
    ]

    # Class methods
    # -------------

    # Instance methods
    # ----------------

    def __init__(self, conn_str:str, user_ns:Dict[str,Any], current:KustoEngine=None, conn_class=None, **options)->Any:
        super(AriaEngine, self).__init__(conn_str, user_ns, current, conn_class, **options)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

from typing import Any, Union, Dict, List
import hashlib
import json
import os
import shutil


from .constants import Constants
from .my_utils import get_valid_filename_with_spaces, adjust_path, convert_to_common_path_obj, json_dumps 
from .kql_response import KqlQueryResponse, KqlSchemaResponse
from .ipython_api import IPythonAPI
from .kql_client import KqlClient
from .kql_engine import KqlEngine


class CacheClient(KqlClient):
    """
    """

    @staticmethod
    def abs_cache_folder(folder_name:str=None, **options)->str:
        root_path = IPythonAPI.get_ipython_root_path(**options)
        cache_folder_name = options.get("cache_folder_name")
        cache_folder_name = f"{Constants.MAGIC_CLASS_NAME_LOWER}/{cache_folder_name}"
        if options.get("temp_folder_location") == "user_dir":
            # app that has a free/tree build server, are not supporting directories athat starts with a dot
            cache_folder_name = f".{cache_folder_name}"
        folder_name_path = f"/{folder_name}" if folder_name else ""
        files_folder = adjust_path(f"{root_path}/{cache_folder_name}{folder_name_path}")
        return files_folder


    @staticmethod
    def remove_cache(folder_name:str=None, **options)->bool:
        cache_folder = CacheClient.abs_cache_folder(folder_name=folder_name, **options)
        if os.path.exists(cache_folder):
            shutil.rmtree(cache_folder)
            return True
        else:
            return False

    @staticmethod
    def list_cache(**options)->List[str]:
        cache_container = CacheClient.abs_cache_folder(**options)
        if os.path.exists(cache_container):
            return os.listdir(cache_container)
        else:
            return []


    @staticmethod
    def create_or_attach_cache(folder_name:str=None, **options)->bool:
        cache_folder = CacheClient.abs_cache_folder(folder_name=folder_name, **options)
        if not os.path.exists(cache_folder):
            os.makedirs(cache_folder)
            return True
        else:
            return False


    def __init__(self, **options)->None:
        """
        File Client constructor.

        Parameters
        ----------
        cluster_folder : str
            folder that contains all the database_folders that contains the query result files
        """

        super(CacheClient, self).__init__()
        self.files_folder = CacheClient.abs_cache_folder(**options)


    @property
    def data_source(self)->str:
        return self.files_folder


    def _get_query_hash_filename(self, query:str)->str:
        lines = [line.replace("\r", "").replace("\t", " ").strip() for line in query.split("\n")]
        q_lines = []
        for line in lines:
            if not line.startswith("//"):
                idx = line.find(" //")
                q_lines.append(line[: idx if idx >= 0 else len(line)])
        return f"q_{hashlib.sha1(bytes(''.join(q_lines), 'utf-8')).hexdigest()}.json"


    def _get_file_path(self, query:str, database_at_cluster:str, cache_folder:str)->str:
        """ 
        get the file name from the query string.

        if query string ends with the '.json' extension it returns the string
        otherwise it computes it from the query
        """

        file_name = query if query.strip().endswith(".json") else self._get_query_hash_filename(query)
        folder_path = self._get_folder_path(database_at_cluster, cache_folder=cache_folder)
        file_path = f"{folder_path}/{file_name}"
        return adjust_path(file_path)


    def _get_folder_path(self, database_at_cluster:str, cache_folder:str=None)->str:
        if "_at_" in database_at_cluster:
            database_at_cluster = "_".join(database_at_cluster.split())
            database_name, cluster_name = database_at_cluster.split("_at_")[:2]

            if not os.path.exists(self.files_folder):
                os.makedirs(self.files_folder)
            folder_path = self.files_folder
            if cache_folder is not None:
                folder_path = adjust_path(f"{folder_path}/{cache_folder}")
                if not os.path.exists(folder_path):
                    os.makedirs(folder_path)
            folder_path = adjust_path(f"{folder_path}/{get_valid_filename_with_spaces(cluster_name)}")
            if not os.path.exists(folder_path):
                os.makedirs(folder_path)
            folder_path = f"{folder_path}/{get_valid_filename_with_spaces(database_name)}"

        else:
            folder_path = database_at_cluster

        folder_path = adjust_path(folder_path)

        if not os.path.exists(folder_path):
            os.makedirs(folder_path)
        return folder_path


    def _get_endpoint_version(self, json_response:Dict[str,Any])->str:
        try:
            tables_num = json_response["Tables"].__len__()  # pylint: disable=W0612
            return "v1"
        except:
            return "v2"


    def execute(self, database_at_cluster:str, query:str, **options)->Union[KqlSchemaResponse,KqlQueryResponse]:
        """
        Executes a query or management command.

        :param str database_at_cluster: name of database and cluster that a folder will be derived that contains all the files with the query results for this specific database.
        :param str query: Query to be executed or a json file with query results.
        """

        file_path = self._get_file_path(query, database_at_cluster, cache_folder=options.get("use_cache"))
        # collect this inormation, in case bug report will be generated
        KqlClient.last_query_info = {
            "request": {
                "file_path": file_path,
            },
            "response": {
                "status_code": 200,
            },
        }
        try:
            str_response = open(file_path, "r").read()
            json_response = json.loads(str_response)
            if query.startswith(".") and json_response.get("tables") is not None:
                return KqlSchemaResponse(json_response)
            else:
                endpoint_version = self._get_endpoint_version(json_response)
                return KqlQueryResponse(json_response, endpoint_version)
                
        except Exception as e:
            # collect this inormation, in case bug report will be generated
            self.last_query_info["response"]["status_code"] = 400  # pylint: disable=unsupported-assignment-operation, unsubscriptable-object
            self.last_query_info["response"]["error"] = str(e)  # pylint: disable=unsupported-assignment-operation, unsubscriptable-object
            raise e


    def save(self, result, engine:KqlEngine, query:str, file_path:str=None, filefolder:str=None, **options)->str:
        """
        Executes a query or management command.

        :param str database_at_cluster: name of database and cluster that a folder will be derived that contains all the files with the query results for this specific database.
        :param str query: Query to be executed.
        """
        if filefolder is not None:
            file_path = f"{filefolder}/{self._get_query_hash_filename(query)}"

        if file_path is not None:
            path_obj = convert_to_common_path_obj(file_path)

            parts = path_obj.get("path").split("/")
            folder_parts = []
            for part in parts[:-1]:
                folder_parts.append(part)
                folder_name = f'{path_obj.get("prefix")}/'.join(folder_parts)
                os_folder_name = adjust_path(folder_name)
                if not os.path.exists(os_folder_name):
                    os.makedirs(os_folder_name)
            file_path = adjust_path(file_path)
            
        else:
            database_friendly_name = engine.get_database_friendly_name()
            cluster_friendly_name = engine.get_cluster_friendly_name()
            file_path = self._get_file_path(query, f"{database_friendly_name}_at_{cluster_friendly_name}", cache_folder=options.get("cache"))
        outfile = open(file_path, "w")
        outfile.write(json_dumps(result.json_response))
        outfile.flush()
        outfile.close()
        return file_path



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

from datetime import datetime, timezone


from .constants import Constants, VisualizationKeys


"""
Splits tabular data in the form of a list of rows into columns;
makes guesses about the role of each column for plotting purposes
(X values, Y values, and text labels).
"""


class Column(list):
    "Store a column of tabular data; record its name and whether it is numeric"

    # Object constructor
    def __init__(self, idx=None, name="", col=None, is_quantity=True, is_datetime=True, **kwargs):
        if col is not None:
            self.is_quantity = col.is_quantity
            self.is_datetime = col.is_datetime
            self.idx = col.idx
            self.name = col.name
        else:
            self.is_quantity = is_quantity
            self.is_datetime = is_datetime
            self.idx = idx
            self.name = name
        super(Column, self).__init__()


class ChartSubTable(dict):
    def __init__(self, col_x=None, col_y=None, name=None, mapping=None, is_descending_sorted=None, col_y_min=None, col_y_max=None,  **kwargs):
        self.is_descending_sorted = is_descending_sorted
        self.col_x = col_x
        self.col_y = col_y
        self.col_y_min = col_y_min
        self.col_y_max = col_y_max
        self.name = name
        super(ChartSubTable, self).__init__()
        if mapping:
            self.update(mapping)


def is_quantity(val) -> bool:
    """Is ``val`` a quantity (int, float, datetime, etc) (not str, bool)?
    
    Relies on presence of __sub__.
    """
    
    return hasattr(val, "__sub__")


def datetime_to_linear_ticks(t: datetime) -> int:
    return (t - datetime(1,1,1,0,0,0,0, tzinfo=timezone.utc)).total_seconds() * Constants.TICK_TO_INT_FACTOR


class ColumnGuesserMixin(object):
    """
    plot: [x, y, y...], y
    pie: ... y
    """

    DATAFRAME_QUNATITY_TYPES = ["int64", "float64", "datetime64[ns]", "timedelta64[ns]", "int32"]
    DATAFRAME_TIME_TYPES = ["datetime64[ns]", "timedelta64[ns]"]

    def _build_chart_sub_tables(self, properties:dict, name=None, x_type="first") -> list:
        self.chart_sub_tables = []
        self._build_columns(name, without_data=True)

        #
        # discover column X index
        #
        x_col_idx = None
        x_col_name = properties.get(VisualizationKeys.X_COLUMN)
        if x_col_name is not None:
            for idx, c in enumerate(self.columns):
                if c.name == x_col_name:
                    x_col_idx = idx
                    break
        elif x_type == "first":
            x_col_idx = 0
        elif x_type == "quantity":
            for idx, c in enumerate(self.columns):
                if c.is_quantity:
                    if not c.is_datetime: 
                        x_col_idx = idx
                        break
                    elif x_col_idx is None:
                        x_col_idx = idx
        elif x_type == "datetime":
            # find first of type datetime
            for idx, c in enumerate(self.columns):
                if c.is_datetime:
                    x_col_idx = idx
                    break
        if x_col_idx is None:
            # this print is not for debug
            print("No valid xcolumn")
            return []

        rows = self

        #
        # discover x direction, and always sort ascending
        #
        is_descending_sorted = None
        if self.columns[x_col_idx].is_quantity and len(rows) >= 2:
            if properties.get(VisualizationKeys.IS_QUERY_SORTED) is True:
                previous_col_value = rows[0][x_col_idx]
                is_descending_sorted = True
                for r in rows[1:]:
                    current_col_value = r[x_col_idx]
                    if previous_col_value < current_col_value:
                        is_descending_sorted = False
                        break
                    previous_col_value = current_col_value

            rows = list(reversed(list(self))) if is_descending_sorted else sorted(self, key=lambda row: row[x_col_idx])

        #
        # create a new unique list of col_x values (keep same order)
        #
        col_x = Column(col=self.columns[x_col_idx])
        for row in rows:
            if row[x_col_idx] not in col_x:
                col_x.append(row[x_col_idx])

        #
        # discover series columns (each combination of values in this columns, is a serie)
        #
        specified_series_columns = properties.get(VisualizationKeys.SERIES) or []
        series_columns = [c for idx, c in enumerate(self.columns) if c.name in specified_series_columns or (idx != x_col_idx and not c.is_quantity)]

        #
        # discover y columns
        #
        y_cols_name = properties.get(VisualizationKeys.Y_COLUMNS)
        if y_cols_name is not None:
            quantity_columns = [c for c in self.columns if c.name in y_cols_name and c.is_quantity]
        else:
            quantity_columns = [c for idx, c in enumerate(self.columns) if idx != x_col_idx and c.is_quantity and c.name not in [s.name for s in series_columns]] 
        if len(quantity_columns) < 1:
            # this print is not for debug
            print("No valid ycolumns")
            return []

        #
        # create chart sub-tables
        # a sub-table for each serie X y-col
        #
        chart_sub_tables_dict = {}
        for row in rows:
            for qcol in quantity_columns:
                if len(series_columns) > 0:
                    sub_table_name = f'{":".join([str(row[col.idx]) for col in series_columns])}:{qcol.name}'
                else:
                    sub_table_name = qcol.name
                chart_sub_table = chart_sub_tables_dict.get(sub_table_name)
                if chart_sub_table is None:
                    chart_sub_table = chart_sub_tables_dict[sub_table_name] = ChartSubTable(
                        name=sub_table_name,
                        col_x=Column(col=self.columns[x_col_idx]),
                        col_y=Column(col=qcol),
                        mapping=dict(zip(col_x, [None for i in range(len(col_x))])),
                        is_descending_sorted=is_descending_sorted,
                    )
                chart_sub_table[row[x_col_idx]] = datetime_to_linear_ticks(row[qcol.idx]) if qcol.is_datetime else row[qcol.idx]
        self.chart_sub_tables = list(chart_sub_tables_dict.values())

        col_y_min = properties.get(VisualizationKeys.Y_MIN)
        col_y_min = col_y_min if is_quantity(col_y_min) else None

        col_y_max = properties.get(VisualizationKeys.Y_MAX)
        col_y_max = col_y_max if is_quantity(col_y_max) else None

        for tab in self.chart_sub_tables:
            tab.col_y_min = col_y_min
            tab.col_y_max = col_y_max
            if tab.col_y_min is None and tab.col_y_max is None:
                pass
            elif tab.col_y_min is not None and tab.col_y_max is not None:
                pass
            elif tab.col_y_min is None:
                try:
                    tab.col_y_min = min(min(filter(lambda x: x is not None, tab.values())), 0) * 1.1
                except:
                    tab.col_y_min = None
            elif tab.col_y_max is None:
                try:
                    tab.col_y_max = max(filter(lambda x: x is not None, tab.values())) * 1.1
                except:
                    tab.col_y_max = None

        return self.chart_sub_tables


    def _build_columns(self, name=None, without_data=False):
        self.x = Column()
        self.ys = []

        rows = self
        if name:
            idx = self.columns_name.index(name)
            rows = sorted(self, key=lambda row: row[idx])  # sort by index

        self.columns = [Column(idx, name) for (idx, name) in enumerate(self.columns_name)]
        if len(self.columns_datafarme_type) > 0:
            for col in self.columns:
                datafarme_type = self.columns_datafarme_type[col.idx]
                col.is_quantity = datafarme_type in self.DATAFRAME_QUNATITY_TYPES
                col.is_datetime = datafarme_type in self.DATAFRAME_TIME_TYPES
            if without_data:
                return

        for row in rows:
            for (col_idx, col_val) in enumerate(row):
                col = self.columns[col_idx]
                if not without_data:
                    col.append(col_val)
                if len(self.columns_datafarme_type) == 0:
                    if (col_val is not None) and (not is_quantity(col_val)):
                        col.is_quantity = False
                        col.is_datetime = False
                    elif not isinstance(col_val, datetime):
                        col.is_datetime = False


    def _get_y(self):
        for idx in range(len(self.columns) - 1, -1, -1):
            if self.columns[idx].is_quantity:
                self.ys.insert(0, self.columns.pop(idx))
                return True


    def _get_x(self):
        for idx in range(len(self.columns)):
            if self.columns[idx].is_quantity:
                self.x = self.columns.pop(idx)
                return True


    def _get_xlabel(self, xlabel_sep=" ", index=None):
        self.xlabels = []
        if self.columns:
            for row_idx in range(len(self.columns[0])):
                self.xlabels.append(xlabel_sep.join(str(c[row_idx]) for c in self.columns))
        self.xlabel = ", ".join(c.name for c in self.columns)


    def _guess_columns(self):
        self._build_columns()
        self._get_y()
        if not self.ys:
            raise AttributeError("No quantitative columns found for chart")


    def build_columns(self, name=None):
        """
        just build the columns.
        if name specified, first sort row by name
        """
        self._build_columns(name)


    def guess_pie_columns(self, xlabel_sep=" "):
        """
        Assigns x, y, and x labels from the data set for a pie chart.
        
        Pie charts simply use the last quantity column as 
        the pie slice size, and everything else as the
        pie slice labels.
        """
        self._guess_columns()
        self._get_xlabel(xlabel_sep)


    def guess_plot_columns(self):
        """
        Assigns ``x`` and ``y`` series from the data set for a plot.
        
        Plots use:
          the rightmost quantity column as a Y series
          optionally, the leftmost quantity column as the X series
          any other quantity columns as additional Y series
        """
        self._guess_columns()
        self._get_x()
        while self._get_y():
            pass
